# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Common cli utilities for transport plugins."""
from functools import partial
import inspect

import click

from aiida.cmdline.params import arguments, options
from aiida.cmdline.params.options.interactive import InteractiveOption
from aiida.cmdline.utils import echo
from aiida.cmdline.utils.decorators import with_dbenv
from aiida.common.exceptions import NotExistent
from aiida.manage import get_manager

TRANSPORT_PARAMS = []


# pylint: disable=unused-argument
def match_comp_transport(ctx, param, computer, transport_type):
    """Check the computer argument against the transport type."""
    if computer.transport_type != transport_type:
        echo.echo_critical(
            f'Computer {computer.label} has transport of type "{computer.transport_type}", not {transport_type}!'
        )
    return computer


@with_dbenv()
def configure_computer_main(computer, user, **kwargs):
    """Configure a computer via the CLI."""
    from aiida import orm

    user = user or orm.User.objects.get_default()

    echo.echo_report(f'Configuring computer {computer.label} for user {user.email}.')
    if user.email != get_manager().get_profile().default_user_email:
        echo.echo_report('Configuring different user, defaults may not be appropriate.')

    computer.configure(user=user, **kwargs)
    echo.echo_success(f'{computer.label} successfully configured for {user.email}')


def common_params(command_func):
    """Decorate a command function with common click parameters for all transport plugins."""
    for param in TRANSPORT_PARAMS.copy().reverse():
        command_func = param(command_func)
    return command_func


def transport_option_default(name, computer):
    """Determine the default value for an auth_param key."""
    transport_cls = computer.get_transport_class()
    suggester_name = f'_get_{name}_suggestion_string'
    members = dict(inspect.getmembers(transport_cls))
    suggester = members.get(suggester_name, None)
    default = None
    if suggester:
        default = suggester(computer)
    else:
        default = transport_cls.auth_options[name].get('default')
    return default


def interactive_default(key, also_non_interactive=False):
    """Create a contextual_default value callback for an auth_param key.

    :param key: the name of the option.
    :param also_non_interactive: indicates whether this option should provide a default also in non-interactive mode. If
        False, the option will raise `MissingParameter` if no explicit value is specified when the command is called in
        non-interactive mode.
    """

    @with_dbenv()
    def get_default(ctx):
        """Determine the default value from the context."""
        from aiida import orm

        if not also_non_interactive and ctx.params['non_interactive']:
            raise click.MissingParameter()

        user = ctx.params.get('user', None) or orm.User.objects.get_default()
        computer = ctx.params.get('computer', None)

        if computer is None:
            return None

        try:
            authinfo = orm.AuthInfo.objects.get(dbcomputer_id=computer.id, aiidauser_id=user.id)
        except NotExistent:
            authinfo = orm.AuthInfo(computer=computer, user=user)

        auth_params = authinfo.get_auth_params()
        suggestion = auth_params.get(key)
        suggestion = suggestion or transport_option_default(key, computer)
        return suggestion

    return get_default


def create_option(name, spec):
    """Create a click option from a name and partial specs as used in transport auth_options."""
    from copy import deepcopy

    spec = deepcopy(spec)
    name_dashed = name.replace('_', '-')
    option_name = f'--{name_dashed}'
    existing_option = spec.pop('option', None)

    if spec.pop('switch', False):
        option_name = '--{name}/--no-{name}'.format(name=name_dashed)

    kwargs = {'cls': InteractiveOption, 'show_default': True}

    non_interactive_default = spec.pop('non_interactive_default', False)
    kwargs['contextual_default'] = interactive_default(name, also_non_interactive=non_interactive_default)

    kwargs.update(spec)

    if existing_option:
        return existing_option(**kwargs)

    return click.option(option_name, **kwargs)


def list_transport_options(transport_type):
    from aiida.plugins import TransportFactory
    options_list = [create_option(*item) for item in TransportFactory(transport_type).auth_options.items()]
    return options_list


def transport_options(transport_type):
    """Decorate a command with all options for a computer configure subcommand for transport_type."""

    def apply_options(func):
        """Decorate the command functionn with the appropriate options for the transport type."""
        options_list = list_transport_options(transport_type)
        options_list.reverse()
        func = arguments.COMPUTER(callback=partial(match_comp_transport, transport_type=transport_type))(func)
        func = options.NON_INTERACTIVE()(func)
        for option in options_list:
            func = option(func)
        func = options.USER()(func)
        func = options.CONFIG_FILE()(func)
        return func

    return apply_options


def create_configure_cmd(transport_type):
    """Create verdi computer configure subcommand for a transport type."""
    help_text = f"""Configure COMPUTER for {transport_type} transport."""

    # pylint: disable=unused-argument
    def transport_configure_command(computer, user, non_interactive, **kwargs):
        """Configure COMPUTER for a type of transport."""
        configure_computer_main(computer, user, **kwargs)

    transport_configure_command.__doc__ = help_text

    return click.command(transport_type)(transport_options(transport_type)(transport_configure_command))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Add indices to `db_dbattribute.tval` and `db_dbnode.uuid`

Revision ID: django_0004
Revises: django_0003

"""
from alembic import op

revision = 'django_0004'
down_revision = 'django_0003'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.execute(
        """
        CREATE INDEX tval_idx_for_daemon
        ON db_dbattribute (tval)
        WHERE ("db_dbattribute"."tval"
        IN ('COMPUTED', 'WITHSCHEDULER', 'TOSUBMIT'))"""
    )
    op.create_index('db_dbnode_uuid_62e0bf98', 'db_dbnode', ['uuid'])
    op.create_index(
        'db_dbnode_uuid_62e0bf98_like',
        'db_dbnode',
        ['uuid'],
        postgresql_using='btree',
        postgresql_ops={'uuid': 'varchar_pattern_ops'},
    )


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0004.')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member,line-too-long
"""Finalise parity of the legacy django branch with the sqlalchemy branch.

1. Remove and recreate all (non-unique) indexes, with standard names and postgresql ops.
2. Remove and recreate all unique constraints, with standard names.
3. Remove and recreate all foreign key constraints, with standard names and other rules.
4. Drop the django specific tables

It is of note that a number of foreign keys were missing comparable `ON DELETE` rules in django.
This is because django does not currently add these rules to the database, but instead tries to handle them on the
Python side, see: https://stackoverflow.com/a/35780859/5033292

Revision ID: django_0050
Revises: django_0049

"""
from alembic import op

from aiida.storage.psql_dos.migrations.utils.parity import synchronize_schemas

revision = 'django_0050'
down_revision = 'django_0049'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    synchronize_schemas(op)

    for tbl_name in (
        'auth_group_permissions', 'auth_permission', 'auth_group', 'django_content_type', 'django_migrations'
    ):
        op.execute(f'DROP TABLE IF EXISTS {tbl_name} CASCADE')


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0050.')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi devel` commands."""
import sys

import click

from aiida import get_profile
from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.utils import decorators, echo


@verdi.group('devel')
def verdi_devel():
    """Commands for developers."""


@verdi_devel.command('check-load-time')
def devel_check_load_time():
    """Check for common indicators that slowdown `verdi`.

    Check for environment properties that negatively affect the responsiveness of the `verdi` command line interface.
    Known pathways that increase load time:

        * the database environment is loaded when it doesn't need to be
        * Unexpected `aiida.*` modules are imported

    If either of these conditions are true, the command will raise a critical error
    """
    from aiida.manage import get_manager

    loaded_aiida_modules = [key for key in sys.modules if key.startswith('aiida.')]
    aiida_modules_str = '\n- '.join(sorted(loaded_aiida_modules))
    echo.echo_info(f'aiida modules loaded:\n- {aiida_modules_str}')

    manager = get_manager()

    if manager.profile_storage_loaded:
        echo.echo_critical('potential `verdi` speed problem: database backend is loaded.')

    allowed = ('aiida.cmdline', 'aiida.common', 'aiida.manage', 'aiida.plugins', 'aiida.restapi')
    for loaded in loaded_aiida_modules:
        if not any(loaded.startswith(mod) for mod in allowed):
            echo.echo_critical(
                f'potential `verdi` speed problem: `{loaded}` module is imported which is not in: {allowed}'
            )

    echo.echo_success('no issues detected')


@verdi_devel.command('check-undesired-imports')
def devel_check_undesired_imports():
    """Check that verdi does not import python modules it shouldn't.

    Note: The blacklist was taken from the list of packages in the 'atomic_tools' extra but can be extended.
    """
    loaded_modules = 0

    for modulename in ['seekpath', 'CifFile', 'ase', 'pymatgen', 'spglib', 'pymysql']:
        if modulename in sys.modules:
            echo.echo_warning(f'Detected loaded module "{modulename}"')
            loaded_modules += 1

    if loaded_modules > 0:
        echo.echo_critical(f'Detected {loaded_modules} unwanted modules')
    echo.echo_success('no issues detected')


@verdi_devel.command('run_daemon')
@decorators.with_dbenv()
def devel_run_daemon():
    """Run a daemon instance in the current interpreter."""
    from aiida.engine.daemon.runner import start_daemon
    start_daemon()


@verdi_devel.command('validate-plugins')
@decorators.with_dbenv()
def devel_validate_plugins():
    """Validate all plugins by checking they can be loaded."""
    from aiida.common.exceptions import EntryPointError
    from aiida.plugins.entry_point import validate_registered_entry_points

    try:
        validate_registered_entry_points()
    except EntryPointError as exception:
        echo.echo_critical(str(exception))

    echo.echo_success('all registered plugins could successfully loaded.')


@verdi_devel.command('run-sql')
@click.argument('sql', type=str)
def devel_run_sql(sql):
    """Run a raw SQL command on the profile database (only available for 'psql_dos' storage)."""
    from sqlalchemy import text

    from aiida.storage.psql_dos.utils import create_sqlalchemy_engine
    assert get_profile().storage_backend == 'psql_dos'
    with create_sqlalchemy_engine(get_profile().storage_config).connect() as connection:
        result = connection.execute(text(sql)).fetchall()

    if isinstance(result, (list, tuple)):
        for row in result:
            echo.echo(str(row))
    else:
        echo.echo(str(result))


@verdi_devel.command('play', hidden=True)
def devel_play():
    """Play the Aida triumphal march by Giuseppe Verdi."""
    import webbrowser
    webbrowser.open_new('http://upload.wikimedia.org/wikipedia/commons/3/32/Triumphal_March_from_Aida.ogg')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""The table models are dynamically generated from the sqlalchemy backend models."""
from __future__ import annotations

from contextlib import contextmanager
from functools import singledispatch
from pathlib import Path
import tempfile
from typing import BinaryIO, Iterable, Iterator, Optional, Sequence, Tuple, Type, cast
from zipfile import ZipFile, is_zipfile

from archive_path import extract_file_in_zip
from sqlalchemy.orm import Session

from aiida.common.exceptions import AiidaException, ClosedStorage, CorruptStorage
from aiida.manage import Profile
from aiida.orm.entities import EntityTypes
from aiida.orm.implementation import StorageBackend
from aiida.repository.backend.abstract import AbstractRepositoryBackend
from aiida.storage.psql_dos.orm import authinfos, comments, computers, entities, groups, logs, nodes, users
from aiida.storage.psql_dos.orm.querybuilder import SqlaQueryBuilder
from aiida.storage.psql_dos.orm.utils import ModelWrapper

from . import models
from .migrator import get_schema_version_head, validate_storage
from .utils import DB_FILENAME, REPO_FOLDER, create_sqla_engine, extract_metadata, read_version


class SqliteZipBackend(StorageBackend):  # pylint: disable=too-many-public-methods
    """A read-only backend for a sqlite/zip format.

    The storage format uses an SQLite database and repository files, within a folder or zipfile.

    The content of the folder/zipfile should be::

        |- metadata.json
        |- db.sqlite3
        |- repo/
            |- hashkey1
            |- hashkey2
            ...

    """

    @classmethod
    def version_head(cls) -> str:
        return get_schema_version_head()

    @staticmethod
    def create_profile(path: str | Path) -> Profile:
        """Create a new profile instance for this backend, from the path to the zip file."""
        profile_name = Path(path).name
        return Profile(
            profile_name, {
                'storage': {
                    'backend': 'sqlite_zip',
                    'config': {
                        'path': str(path)
                    }
                },
                'process_control': {
                    'backend': 'null',
                    'config': {}
                }
            }
        )

    @classmethod
    def version_profile(cls, profile: Profile) -> Optional[str]:
        return read_version(profile.storage_config['path'], search_limit=None)

    @classmethod
    def migrate(cls, profile: Profile):
        raise NotImplementedError('use the migrate function directly.')

    def __init__(self, profile: Profile):
        super().__init__(profile)
        self._path = Path(profile.storage_config['path'])
        validate_storage(self._path)
        # lazy open the archive zipfile and extract the database file
        self._db_file: Optional[Path] = None
        self._session: Optional[Session] = None
        self._repo: Optional[_RoBackendRepository] = None
        self._closed = False

    def __str__(self) -> str:
        state = 'closed' if self.is_closed else 'open'
        return f'SqliteZip storage (read-only) [{state}] @ {self._path}'

    @property
    def is_closed(self) -> bool:
        return self._closed

    def close(self):
        """Close the backend"""
        if self._session:
            self._session.close()
        if self._db_file and self._db_file.exists():
            self._db_file.unlink()
        if self._repo:
            self._repo.close()
        self._session = None
        self._db_file = None
        self._repo = None
        self._closed = True

    def get_session(self) -> Session:
        """Return an SQLAlchemy session."""
        if self._closed:
            raise ClosedStorage(str(self))
        if self._session is None:
            if is_zipfile(self._path):
                _, path = tempfile.mkstemp()
                db_file = self._db_file = Path(path)
                with db_file.open('wb') as handle:
                    try:
                        extract_file_in_zip(self._path, DB_FILENAME, handle, search_limit=4)
                    except Exception as exc:
                        raise CorruptStorage(f'database could not be read: {exc}') from exc
            else:
                db_file = self._path / DB_FILENAME
                if not db_file.exists():
                    raise CorruptStorage(f'database could not be read: non-existent {db_file}')
            self._session = Session(create_sqla_engine(db_file))
        return self._session

    def get_repository(self) -> '_RoBackendRepository':
        if self._closed:
            raise ClosedStorage(str(self))
        if self._repo is None:
            if is_zipfile(self._path):
                self._repo = ZipfileBackendRepository(self._path)
            elif (self._path / REPO_FOLDER).exists():
                self._repo = FolderBackendRepository(self._path / REPO_FOLDER)
            else:
                raise CorruptStorage(f'repository could not be read: non-existent {self._path / REPO_FOLDER}')
        return self._repo

    def query(self) -> 'SqliteBackendQueryBuilder':
        return SqliteBackendQueryBuilder(self)

    def get_backend_entity(self, res):  # pylint: disable=no-self-use
        """Return the backend entity that corresponds to the given Model instance."""
        klass = get_backend_entity(res)
        return klass(self, res)

    @property
    def authinfos(self):
        return create_backend_collection(
            authinfos.SqlaAuthInfoCollection, self, authinfos.SqlaAuthInfo, models.DbAuthInfo
        )

    @property
    def comments(self):
        return create_backend_collection(comments.SqlaCommentCollection, self, comments.SqlaComment, models.DbComment)

    @property
    def computers(self):
        return create_backend_collection(
            computers.SqlaComputerCollection, self, computers.SqlaComputer, models.DbComputer
        )

    @property
    def groups(self):
        return create_backend_collection(groups.SqlaGroupCollection, self, groups.SqlaGroup, models.DbGroup)

    @property
    def logs(self):
        return create_backend_collection(logs.SqlaLogCollection, self, logs.SqlaLog, models.DbLog)

    @property
    def nodes(self):
        return create_backend_collection(nodes.SqlaNodeCollection, self, nodes.SqlaNode, models.DbNode)

    @property
    def users(self):
        return create_backend_collection(users.SqlaUserCollection, self, users.SqlaUser, models.DbUser)

    def _clear(self, recreate_user: bool = True) -> None:
        raise ReadOnlyError()

    def transaction(self):
        raise ReadOnlyError()

    @property
    def in_transaction(self) -> bool:
        return False

    def bulk_insert(self, entity_type: EntityTypes, rows: list[dict], allow_defaults: bool = False) -> list[int]:
        raise ReadOnlyError()

    def bulk_update(self, entity_type: EntityTypes, rows: list[dict]) -> None:
        raise ReadOnlyError()

    def delete_nodes_and_connections(self, pks_to_delete: Sequence[int]):
        raise ReadOnlyError()

    def get_global_variable(self, key: str):
        raise NotImplementedError

    def set_global_variable(self, key: str, value, description: Optional[str] = None, overwrite=True) -> None:
        raise ReadOnlyError()

    def maintain(self, dry_run: bool = False, live: bool = True, **kwargs) -> None:
        raise NotImplementedError

    def get_info(self, detailed: bool = False) -> dict:
        # since extracting the database file is expensive, we only do it if detailed is True
        results = {'metadata': extract_metadata(self._path)}
        if detailed:
            results.update(super().get_info(detailed=detailed))
            results['repository'] = self.get_repository().get_info(detailed)
        return results


class ReadOnlyError(AiidaException):
    """Raised when a write operation is called on a read-only archive."""

    def __init__(self, msg='sqlite_zip storage is read-only'):  # pylint: disable=useless-super-delegation
        super().__init__(msg)


class _RoBackendRepository(AbstractRepositoryBackend):  # pylint: disable=abstract-method
    """A backend abstract for a read-only folder or zip file."""

    def __init__(self, path: str | Path):
        """Initialise the repository backend.

        :param path: the path to the zip file
        """
        self._path = Path(path)
        self._closed = False

    def close(self) -> None:
        """Close the repository."""
        self._closed = True

    @property
    def uuid(self) -> Optional[str]:
        return None

    @property
    def key_format(self) -> Optional[str]:
        return 'sha256'

    def initialise(self, **kwargs) -> None:
        pass

    @property
    def is_initialised(self) -> bool:
        return True

    def erase(self) -> None:
        raise ReadOnlyError()

    def _put_object_from_filelike(self, handle: BinaryIO) -> str:
        raise ReadOnlyError()

    def has_objects(self, keys: list[str]) -> list[bool]:
        return [self.has_object(key) for key in keys]

    def iter_object_streams(self, keys: list[str]) -> Iterator[Tuple[str, BinaryIO]]:
        for key in keys:
            with self.open(key) as handle:  # pylint: disable=not-context-manager
                yield key, handle

    def delete_objects(self, keys: list[str]) -> None:
        raise ReadOnlyError()

    def get_object_hash(self, key: str) -> str:
        return key

    def maintain(self, dry_run: bool = False, live: bool = True, **kwargs) -> None:
        pass

    def get_info(self, detailed: bool = False, **kwargs) -> dict:
        return {'objects': {'count': len(list(self.list_objects()))}}


class ZipfileBackendRepository(_RoBackendRepository):
    """A read-only backend for a zip file.

    The zip file should contain repository files with the key format: ``repo/<sha256 hash>``,
    i.e. files named by the sha256 hash of the file contents, inside a ``repo`` directory.
    """

    def __init__(self, path: str | Path):
        super().__init__(path)
        self._folder = REPO_FOLDER
        self.__zipfile: None | ZipFile = None

    def close(self) -> None:
        if self._zipfile:
            self._zipfile.close()
        super().close()

    @property
    def _zipfile(self) -> ZipFile:
        """Return the open zip file."""
        if self._closed:
            raise ClosedStorage(f'repository is closed: {self._path}')
        if self.__zipfile is None:
            try:
                self.__zipfile = ZipFile(self._path, mode='r')  # pylint: disable=consider-using-with
            except Exception as exc:
                raise CorruptStorage(f'repository could not be read {self._path}: {exc}') from exc
        return self.__zipfile

    def has_object(self, key: str) -> bool:
        try:
            self._zipfile.getinfo(f'{self._folder}/{key}')
        except KeyError:
            return False
        return True

    def list_objects(self) -> Iterable[str]:
        prefix = f'{self._folder}/'
        prefix_len = len(prefix)
        for name in self._zipfile.namelist():
            if name.startswith(prefix) and name[prefix_len:]:
                yield name[prefix_len:]

    @contextmanager
    def open(self, key: str) -> Iterator[BinaryIO]:
        try:
            handle = self._zipfile.open(f'{self._folder}/{key}')
            yield cast(BinaryIO, handle)
        except KeyError:
            raise FileNotFoundError(f'object with key `{key}` does not exist.')
        finally:
            handle.close()


class FolderBackendRepository(_RoBackendRepository):
    """A read-only backend for a folder.

    The folder should contain repository files, named by the sha256 hash of the file contents.
    """

    def has_object(self, key: str) -> bool:
        return self._path.joinpath(key).is_file()

    def list_objects(self) -> Iterable[str]:
        for subpath in self._path.iterdir():
            if subpath.is_file():
                yield subpath.name

    @contextmanager
    def open(self, key: str) -> Iterator[BinaryIO]:
        if not self._path.joinpath(key).is_file():
            raise FileNotFoundError(f'object with key `{key}` does not exist.')
        with self._path.joinpath(key).open('rb') as handle:
            yield handle


class SqliteBackendQueryBuilder(SqlaQueryBuilder):
    """Archive query builder"""

    @property
    def Node(self):
        return models.DbNode

    @property
    def Link(self):
        return models.DbLink

    @property
    def Computer(self):
        return models.DbComputer

    @property
    def User(self):
        return models.DbUser

    @property
    def Group(self):
        return models.DbGroup

    @property
    def AuthInfo(self):
        return models.DbAuthInfo

    @property
    def Comment(self):
        return models.DbComment

    @property
    def Log(self):
        return models.DbLog

    @property
    def table_groups_nodes(self):
        return models.DbGroupNodes.__table__  # type: ignore[attr-defined] # pylint: disable=no-member


def create_backend_cls(base_class, model_cls):
    """Create an archive backend class for the given model class."""

    class ReadOnlyEntityBackend(base_class):  # type: ignore
        """Backend class for the read-only archive."""

        MODEL_CLASS = model_cls

        def __init__(self, _backend, model):
            """Initialise the backend entity."""
            self._backend = _backend
            self._model = ModelWrapper(model, _backend)

        @property
        def model(self) -> ModelWrapper:
            """Return an ORM model that correctly updates and flushes the data model when getting or setting a field."""
            return self._model

        @property
        def bare_model(self):
            """Return the underlying SQLAlchemy ORM model for this entity."""
            return self.model._model  # pylint: disable=protected-access

        @classmethod
        def from_dbmodel(cls, model, _backend):
            return cls(_backend, model)

        @property
        def is_stored(self):
            return True

        def store(self):  # pylint: disable=no-self-use
            raise ReadOnlyError()

    return ReadOnlyEntityBackend


def create_backend_collection(cls, _backend, entity_cls, model):
    collection = cls(_backend)
    new_cls = create_backend_cls(entity_cls, model)
    collection.ENTITY_CLASS = new_cls
    return collection


@singledispatch
def get_backend_entity(dbmodel) -> Type[entities.SqlaModelEntity]:  # pylint: disable=unused-argument
    raise TypeError(f'Cannot get backend entity for {dbmodel}')


@get_backend_entity.register(models.DbAuthInfo)  # type: ignore[call-overload]
def _(dbmodel):
    return create_backend_cls(authinfos.SqlaAuthInfo, dbmodel.__class__)


@get_backend_entity.register(models.DbComment)  # type: ignore[call-overload]
def _(dbmodel):
    return create_backend_cls(comments.SqlaComment, dbmodel.__class__)


@get_backend_entity.register(models.DbComputer)  # type: ignore[call-overload]
def _(dbmodel):
    return create_backend_cls(computers.SqlaComputer, dbmodel.__class__)


@get_backend_entity.register(models.DbGroup)  # type: ignore[call-overload]
def _(dbmodel):
    return create_backend_cls(groups.SqlaGroup, dbmodel.__class__)


@get_backend_entity.register(models.DbLog)  # type: ignore[call-overload]
def _(dbmodel):
    return create_backend_cls(logs.SqlaLog, dbmodel.__class__)


@get_backend_entity.register(models.DbNode)  # type: ignore[call-overload]
def _(dbmodel):
    return create_backend_cls(nodes.SqlaNode, dbmodel.__class__)


@get_backend_entity.register(models.DbUser)  # type: ignore[call-overload]
def _(dbmodel):
    return create_backend_cls(users.SqlaUser, dbmodel.__class__)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Methods and definitions of migrations for the configuration file of an AiiDA instance."""

# AUTO-GENERATED

# yapf: disable
# pylint: disable=wildcard-import

from .migrations import *

__all__ = (
    'CURRENT_CONFIG_VERSION',
    'MIGRATIONS',
    'OLDEST_COMPATIBLE_CONFIG_VERSION',
    'check_and_migrate_config',
    'config_needs_migrating',
    'downgrade_config',
    'get_current_version',
    'upgrade_config',
)

# yapf: enable



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Add `db_dblink.type` field, and remove link field uniqueness constraints

Revision ID: django_0003
Revises: django_0002

"""
from alembic import op
import sqlalchemy as sa

from aiida.storage.psql_dos.migrations.utils import ReflectMigrations

revision = 'django_0003'
down_revision = 'django_0002'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.add_column('db_dblink', sa.Column('type', sa.VARCHAR(length=255), nullable=False, server_default=''))
    op.alter_column('db_dblink', 'type', server_default=None)
    op.create_index('db_dblink_type_229f212b', 'db_dblink', ['type'])
    op.create_index(
        'db_dblink_type_229f212b_like',
        'db_dblink',
        ['type'],
        postgresql_using='btree',
        postgresql_ops={'type': 'varchar_pattern_ops'},
    )
    reflect = ReflectMigrations(op)
    reflect.drop_unique_constraints('db_dblink', ['input_id', 'output_id'])
    reflect.drop_unique_constraints('db_dblink', ['output_id', 'label'])


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0003.')



# -*- coding: utf-8 -*-
"""Module for file repository backend implementations."""

# AUTO-GENERATED

# yapf: disable
# pylint: disable=wildcard-import

from .abstract import *
from .disk_object_store import *
from .sandbox import *

__all__ = (
    'AbstractRepositoryBackend',
    'DiskObjectStoreRepositoryBackend',
    'SandboxRepositoryBackend',
)

# yapf: enable



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module that defines the exceptions thrown by AiiDA's archive module.

Note: In order to not override the built-in `ImportError`,
    both `ImportError` and `ExportError` are prefixed with `Archive`.
"""

from aiida.common.exceptions import AiidaException

__all__ = (
    'ExportImportException',
    'ArchiveExportError',
    'ExportValidationError',
    'ArchiveImportError',
    'ImportValidationError',
    'ImportUniquenessError',
    'ImportTestRun',
)


class ExportImportException(AiidaException):
    """Base class for all AiiDA export/import module exceptions."""


class ArchiveExportError(ExportImportException):
    """Base class for all AiiDA export exceptions."""


class ExportValidationError(ArchiveExportError):
    """Raised when validation fails during export, e.g. for non-sealed ``ProcessNode`` s."""


class ArchiveImportError(ExportImportException):
    """Base class for all AiiDA import exceptions."""


class ImportUniquenessError(ArchiveImportError):
    """Raised when the user tries to violate a uniqueness constraint.

    Similar to :py:class:`~aiida.common.exceptions.UniquenessError`.
    """


class ImportValidationError(ArchiveImportError):
    """Raised when validation fails during import, e.g. for parameter types and values."""


class ImportTestRun(ArchiveImportError):
    """Raised during an import, before the transaction is commited."""



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi profile` command."""
import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import arguments, options
from aiida.cmdline.utils import defaults, echo
from aiida.common import exceptions
from aiida.manage.configuration import get_config


@verdi.group('profile')
def verdi_profile():
    """Inspect and manage the configured profiles."""


@verdi_profile.command('list')
def profile_list():
    """Display a list of all available profiles."""
    try:
        config = get_config()
    except (exceptions.MissingConfigurationError, exceptions.ConfigurationError) as exception:
        # This can happen for a fresh install and the `verdi setup` has not yet been run. In this case it is still nice
        # to be able to see the configuration directory, for instance for those who have set `AIIDA_PATH`. This way
        # they can at least verify that it is correctly set.
        from aiida.manage.configuration.settings import AIIDA_CONFIG_FOLDER
        echo.echo_report(f'configuration folder: {AIIDA_CONFIG_FOLDER}')
        echo.echo_critical(str(exception))
    else:
        echo.echo_report(f'configuration folder: {config.dirpath}')

    if not config.profiles:
        echo.echo_warning('no profiles configured: run `verdi setup` to create one')
    else:
        sort = lambda profile: profile.name
        highlight = lambda profile: profile.name == config.default_profile_name
        echo.echo_formatted_list(config.profiles, ['name'], sort=sort, highlight=highlight)


def _strip_private_keys(dct: dict):
    """Remove private keys (starting `_`) from the dictionary."""
    return {
        key: _strip_private_keys(value) if isinstance(value, dict) else value
        for key, value in dct.items()
        if not key.startswith('_')
    }


@verdi_profile.command('show')
@arguments.PROFILE(default=defaults.get_default_profile)
def profile_show(profile):
    """Show details for a profile."""

    if profile is None:
        echo.echo_critical('no profile to show')

    echo.echo_report(f'Profile: {profile.name}')
    config = _strip_private_keys(profile.dictionary)
    echo.echo_dictionary(config, fmt='yaml')


@verdi_profile.command('setdefault')
@arguments.PROFILE(required=True, default=None)
def profile_setdefault(profile):
    """Set a profile as the default one."""
    try:
        config = get_config()
    except (exceptions.MissingConfigurationError, exceptions.ConfigurationError) as exception:
        echo.echo_critical(str(exception))

    config.set_default_profile(profile.name, overwrite=True).store()
    echo.echo_success(f'{profile.name} set as default profile')


@verdi_profile.command('delete')
@options.FORCE(help='to skip questions and warnings about loss of data')
@click.option(
    '--include-config/--skip-config',
    default=True,
    show_default=True,
    help='Include deletion of entry in configuration file.'
)
@click.option(
    '--include-db/--skip-db',
    'include_database',
    default=True,
    show_default=True,
    help='Include deletion of associated database.'
)
@click.option(
    '--include-db-user/--skip-db-user',
    'include_database_user',
    default=False,
    show_default=True,
    help='Include deletion of associated database user.'
)
@click.option(
    '--include-repository/--skip-repository',
    default=True,
    show_default=True,
    help='Include deletion of associated file repository.'
)
@arguments.PROFILES(required=True)
def profile_delete(force, include_config, include_database, include_database_user, include_repository, profiles):
    """Delete one or more profiles.

    The PROFILES argument takes one or multiple profile names that will be deleted. Deletion here means that the profile
    will be removed including its file repository and database. The various options can be used to control which parts
    of the profile are deleted.
    """
    if not include_config:
        echo.echo_deprecated('the `--skip-config` option is deprecated and is no longer respected.')

    for profile in profiles:

        includes = {
            'database': include_database,
            'database user': include_database_user,
            'file repository': include_repository
        }

        if not all(includes.values()):
            excludes = [label for label, value in includes.items() if not value]
            message_suffix = f' excluding: {", ".join(excludes)}.'
        else:
            message_suffix = '.'

        echo.echo_warning(f'deleting profile `{profile.name}`{message_suffix}')
        echo.echo_warning('this operation cannot be undone, ', nl=False)

        if not force and not click.confirm('are you sure you want to continue?'):
            echo.echo_report(f'deleting of `{profile.name} cancelled.')
            continue

        get_config().delete_profile(
            profile.name,
            include_database=include_database,
            include_database_user=include_database_user,
            include_repository=include_repository
        )
        echo.echo_success(f'profile `{profile.name}` was deleted{message_suffix}.')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""This migration adds uniqueness constraint to the UUID column.

This migration corresponds to the 0024_dblog_update Django migration.

Revision ID: 375c2db70663
Revises: ea2f50e7f615
Create Date: 2019-01-30 20:26:16.550071

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '375c2db70663'
down_revision = 'ea2f50e7f615'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.create_unique_constraint('db_dblog_uuid_key', 'db_dblog', ['uuid'])


def downgrade():
    """Migrations for the downgrade."""
    op.drop_constraint('db_dblog_uuid_key', 'db_dblog')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for the ORM user class."""
from typing import TYPE_CHECKING, Optional, Tuple, Type

from aiida.common import exceptions
from aiida.common.lang import classproperty
from aiida.manage import get_manager

from . import entities

if TYPE_CHECKING:
    from aiida.orm.implementation import BackendUser, StorageBackend

__all__ = ('User',)


class UserCollection(entities.Collection['User']):
    """The collection of users stored in a backend."""

    @staticmethod
    def _entity_base_cls() -> Type['User']:
        return User

    def __init__(self, entity_class: Type['User'], backend: Optional['StorageBackend'] = None) -> None:
        super().__init__(entity_class=entity_class, backend=backend)
        self._default_user: Optional[User] = None

    def get_or_create(self, email: str, **kwargs) -> Tuple[bool, 'User']:
        """Get the existing user with a given email address or create an unstored one

        :param kwargs: The properties of the user to get or create
        :return: The corresponding user object
        :raises: :class:`aiida.common.exceptions.MultipleObjectsError`,
            :class:`aiida.common.exceptions.NotExistent`
        """
        try:
            return False, self.get(email=email)
        except exceptions.NotExistent:
            return True, User(backend=self.backend, email=email, **kwargs)

    def get_default(self) -> Optional['User']:
        """Get the current default user"""
        if self._default_user is None:
            email = self.backend.profile.default_user_email
            if not email:
                self._default_user = None

            try:
                self._default_user = self.get(email=email)
            except (exceptions.MultipleObjectsError, exceptions.NotExistent):
                self._default_user = None

        return self._default_user

    def reset(self) -> None:
        """
        Reset internal caches (default user).
        """
        self._default_user = None


class User(entities.Entity['BackendUser']):
    """AiiDA User"""

    Collection = UserCollection

    @classproperty
    def objects(cls: Type['User']) -> UserCollection:  # type: ignore[misc] # pylint: disable=no-self-argument
        return UserCollection.get_cached(cls, get_manager().get_profile_storage())

    def __init__(
        self,
        email: str,
        first_name: str = '',
        last_name: str = '',
        institution: str = '',
        backend: Optional['StorageBackend'] = None
    ):
        """Create a new `User`."""
        # pylint: disable=too-many-arguments
        backend = backend or get_manager().get_profile_storage()
        email = self.normalize_email(email)
        backend_entity = backend.users.create(email, first_name, last_name, institution)
        super().__init__(backend_entity)

    def __str__(self) -> str:
        return self.email

    @staticmethod
    def normalize_email(email: str) -> str:
        """Normalize the address by lowercasing the domain part of the email address (taken from Django)."""
        email = email or ''
        try:
            email_name, domain_part = email.strip().rsplit('@', 1)
        except ValueError:
            pass
        else:
            email = '@'.join([email_name, domain_part.lower()])
        return email

    @property
    def email(self) -> str:
        return self._backend_entity.email

    @email.setter
    def email(self, email: str) -> None:
        self._backend_entity.email = email

    @property
    def first_name(self) -> str:
        return self._backend_entity.first_name

    @first_name.setter
    def first_name(self, first_name: str) -> None:
        self._backend_entity.first_name = first_name

    @property
    def last_name(self) -> str:
        return self._backend_entity.last_name

    @last_name.setter
    def last_name(self, last_name: str) -> None:
        self._backend_entity.last_name = last_name

    @property
    def institution(self) -> str:
        return self._backend_entity.institution

    @institution.setter
    def institution(self, institution: str) -> None:
        self._backend_entity.institution = institution

    def get_full_name(self) -> str:
        """
        Return the user full name

        :return: the user full name
        """
        if self.first_name and self.last_name:
            full_name = f'{self.first_name} {self.last_name} ({self.email})'
        elif self.first_name:
            full_name = f'{self.first_name} ({self.email})'
        elif self.last_name:
            full_name = f'{self.last_name} ({self.email})'
        else:
            full_name = f'{self.email}'

        return full_name

    def get_short_name(self) -> str:
        """
        Return the user short name (typically, this returns the email)

        :return: The short name
        """
        return self.email

    @property
    def uuid(self) -> None:
        """
        For now users do not have UUIDs so always return None
        """
        return None



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Convenience functions for logging output from ``verdi`` commands."""
import collections
import enum
import json
import sys

import click
import yaml

from aiida.common.log import AIIDA_LOGGER

CMDLINE_LOGGER = AIIDA_LOGGER.getChild('cmdline')

__all__ = ('echo_report', 'echo_info', 'echo_success', 'echo_warning', 'echo_error', 'echo_critical', 'echo_dictionary')


class ExitCode(enum.IntEnum):
    """Exit codes for the verdi command line."""
    CRITICAL = 1
    DEPRECATED = 80
    UNKNOWN = 99
    SUCCESS = 0


COLORS = {
    'success': 'green',
    'highlight': 'green',
    'debug': 'white',
    'info': 'blue',
    'report': 'blue',
    'warning': 'bright_yellow',
    'error': 'red',
    'critical': 'red',
    'deprecated': 'red',
}


def echo(message: str, fg: str = None, bold: bool = False, nl: bool = True, err: bool = False) -> None:
    """Log a message to the cmdline logger.

    .. note:: The message will be logged at the ``REPORT`` level but always without the log level prefix.

    :param message: the message to log.
    :param fg: if provided this will become the foreground color.
    :param bold: whether to print the messaformat bold.
    :param nl: whether to print a newlineaddhe end of the message.
    :param err: whether to log to stderr.
    """
    message = click.style(message, fg=fg, bold=bold)
    CMDLINE_LOGGER.report(message, extra=dict(nl=nl, err=err, prefix=False))


def echo_debug(message: str, bold: bool = False, nl: bool = True, err: bool = False, prefix: bool = True) -> None:
    """Log a debug message to the cmdline logger.

    :param message: the message to log.
    :param bold: whether to format the message in bold.
    :param nl: whether to add a newline at the end of the message.
    :param err: whether to log to stderr.
    :param prefix: whether the message should be prefixed with a colored version of the log level.
    """
    message = click.style(message, bold=bold)
    CMDLINE_LOGGER.debug(message, extra=dict(nl=nl, err=err, prefix=prefix))


def echo_info(message: str, bold: bool = False, nl: bool = True, err: bool = False, prefix: bool = True) -> None:
    """Log an info message to the cmdline logger.

    :param message: the message to log.
    :param bold: whether to format the message in bold.
    :param nl: whether to add a newline at the end of the message.
    :param err: whether to log to stderr.
    :param prefix: whether the message should be prefixed with a colored version of the log level.
    """
    message = click.style(message, bold=bold)
    CMDLINE_LOGGER.info(message, extra=dict(nl=nl, err=err, prefix=prefix))


def echo_report(message: str, bold: bool = False, nl: bool = True, err: bool = False, prefix: bool = True) -> None:
    """Log an report message to the cmdline logger.

    :param message: the message to log.
    :param bold: whether to format the message in bold.
    :param nl: whether to add a newline at the end of the message.
    :param err: whether to log to stderr.
    :param prefix: whether the message should be prefixed with a colored version of the log level.
    """
    message = click.style(message, bold=bold)
    CMDLINE_LOGGER.report(message, extra=dict(nl=nl, err=err, prefix=prefix))


def echo_success(message: str, bold: bool = False, nl: bool = True, err: bool = False, prefix: bool = True) -> None:
    """Log a success message to the cmdline logger.

    .. note:: The message will be logged at the ``REPORT`` level and always with the ``Success:`` prefix.

    :param message: the message to log.
    :param bold: whether to format the message in bold.
    :param nl: whether to add a newline at the end of the message.
    :param err: whether to log to stderr.
    :param prefix: whether the message should be prefixed with a colored version of the log level.
    """
    message = click.style(message, bold=bold)

    if prefix:
        message = click.style('Success: ', bold=True, fg=COLORS['success']) + message

    CMDLINE_LOGGER.report(message, extra=dict(nl=nl, err=err, prefix=False))


def echo_warning(message: str, bold: bool = False, nl: bool = True, err: bool = False, prefix: bool = True) -> None:
    """Log a warning message to the cmdline logger.

    :param message: the message to log.
    :param bold: whether to format the message in bold.
    :param nl: whether to add a newline at the end of the message.
    :param err: whether to log to stderr.
    :param prefix: whether the message should be prefixed with a colored version of the log level.
    """
    message = click.style(message, bold=bold)
    CMDLINE_LOGGER.warning(message, extra=dict(nl=nl, err=err, prefix=prefix))


def echo_error(message: str, bold: bool = False, nl: bool = True, err: bool = True, prefix: bool = True) -> None:
    """Log an error message to the cmdline logger.

    :param message: the message to log.
    :param bold: whether to format the message in bold.
    :param nl: whether to add a newline at the end of the message.
    :param err: whether to log to stderr.
    :param prefix: whether the message should be prefixed with a colored version of the log level.
    """
    message = click.style(message, bold=bold)
    CMDLINE_LOGGER.error(message, extra=dict(nl=nl, err=err, prefix=prefix))


def echo_critical(message: str, bold: bool = False, nl: bool = True, err: bool = True, prefix: bool = True) -> None:
    """Log a critical error message to the cmdline logger and exit with ``exit_status``.

    This should be used to print messages for errors that cannot be recovered from and so the script should be directly
    terminated with a non-zero exit status to indicate that the command failed.

    :param message: the message to log.
    :param bold: whether to format the message in bold.
    :param nl: whether to add a newline at the end of the message.
    :param err: whether to log to stderr.
    :param prefix: whether the message should be prefixed with a colored version of the log level.
    """
    message = click.style(message, bold=bold)
    CMDLINE_LOGGER.critical(message, extra=dict(nl=nl, err=err, prefix=prefix))
    sys.exit(ExitCode.CRITICAL)


def echo_deprecated(message: str, bold: bool = False, nl: bool = True, err: bool = True, exit: bool = False) -> None:
    """Log an error message to the cmdline logger, prefixed with 'Deprecated:' exiting with the given ``exit_status``.

    This should be used to indicate deprecated commands.

    :param message: the message to log.
    :param bold: whether to format the message in bold.
    :param nl: whether to add a newline at the end of the message.
    :param err: whether to log to stderr.
    :param exit: whether to exit after printing the message
    """
    # pylint: disable=redefined-builtin
    prefix = click.style('Deprecated: ', fg=COLORS['deprecated'], bold=True)
    echo_warning(prefix + message, bold=bold, nl=nl, err=err, prefix=False)

    if exit:
        sys.exit(ExitCode.DEPRECATED)


def echo_formatted_list(collection, attributes, sort=None, highlight=None, hide=None):
    """Log a collection of entries as a formatted list, one entry per line.

    :param collection: a list of objects
    :param attributes: a list of attributes to log for each entry in the collection
    :param sort: optional lambda to sort the collection
    :param highlight: optional lambda to highlight an entry in the collection if it returns True
    :param hide: optional lambda to skip an entry if it returns True
    """
    if sort:
        entries = sorted(collection, key=sort)
    else:
        entries = collection

    template = f"{{symbol}}{' {}' * len(attributes)}"

    for entry in entries:
        if hide and hide(entry):
            continue

        values = [getattr(entry, attribute) for attribute in attributes]
        if highlight and highlight(entry):
            echo(click.style(template.format(symbol='*', *values), fg=COLORS['highlight']))
        else:
            echo(click.style(template.format(symbol=' ', *values)))


def _format_dictionary_json_date(dictionary, sort_keys=True):
    """Return a dictionary formatted as a string using the json format and converting dates to strings."""

    def default_jsondump(data):
        """Function needed to decode datetimes, that would otherwise not be JSON-decodable."""
        import datetime

        from aiida.common import timezone

        if isinstance(data, datetime.datetime):
            return timezone.localtime(data).strftime('%Y-%m-%dT%H:%M:%S.%f%z')

        raise TypeError(f'{repr(data)} is not JSON serializable')

    return json.dumps(dictionary, indent=4, sort_keys=sort_keys, default=default_jsondump)


def _format_yaml(dictionary, sort_keys=True):
    """Return a dictionary formatted as a string using the YAML format."""
    return yaml.dump(dictionary, sort_keys=sort_keys)


def _format_yaml_expanded(dictionary, sort_keys=True):
    """Return a dictionary formatted as a string using the expanded YAML format."""
    return yaml.dump(dictionary, sort_keys=sort_keys, default_flow_style=False)


VALID_DICT_FORMATS_MAPPING = collections.OrderedDict(
    (('json+date', _format_dictionary_json_date), ('yaml', _format_yaml), ('yaml_expanded', _format_yaml_expanded))
)


def echo_dictionary(dictionary, fmt='json+date', sort_keys=True):
    """Log the given dictionary to stdout in the given format

    :param dictionary: the dictionary
    :param fmt: the format to use for printing
    :param sort_keys: Whether to automatically sort keys
    """
    try:
        format_function = VALID_DICT_FORMATS_MAPPING[fmt]
    except KeyError:
        formats = ', '.join(VALID_DICT_FORMATS_MAPPING.keys())
        raise ValueError(f'Unrecognised printing format. Valid formats are: {formats}')

    echo(format_function(dictionary, sort_keys=sort_keys))


def is_stdout_redirected():
    """Determines if the standard output is redirected.

    For cases where the standard output is redirected and you want to
    inform the user without messing up the output. Example::

        echo.echo_info("Found {} results".format(qb.count()), err=echo.is_stdout_redirected)
        echo.echo(tabulate.tabulate(qb.all()))
    """
    # pylint: disable=no-member
    return not sys.stdout.isatty()



# -*- coding: utf-8 -*-
"""Implementation of the ``AbstractRepositoryBackend`` using a sandbox folder on disk as the backend."""
import contextlib
import os
import shutil
from typing import BinaryIO, Iterable, Iterator, List, Optional, Tuple
import uuid

from aiida.common.folders import SandboxFolder

from .abstract import AbstractRepositoryBackend

__all__ = ('SandboxRepositoryBackend',)


class SandboxRepositoryBackend(AbstractRepositoryBackend):
    """Implementation of the ``AbstractRepositoryBackend`` using a sandbox folder on disk as the backend."""

    def __init__(self):
        self._sandbox: Optional[SandboxFolder] = None

    def __str__(self) -> str:
        """Return the string representation of this repository."""
        if self.is_initialised:
            return f'SandboxRepository: {self._sandbox.abspath if self._sandbox else "null"}'
        return 'SandboxRepository: <uninitialised>'

    def __del__(self):
        """Delete the entire sandbox folder if it was instantiated and still exists."""
        self.erase()

    @property
    def uuid(self) -> Optional[str]:
        """Return the unique identifier of the repository.

        .. note:: A sandbox folder does not have the concept of a unique identifier and so always returns ``None``.
        """
        return None

    @property
    def key_format(self) -> Optional[str]:
        return 'uuid4'

    def initialise(self, **kwargs) -> None:
        """Initialise the repository if it hasn't already been initialised.

        :param kwargs: parameters for the initialisation.
        """
        # Merely calling the property will cause the sandbox folder to be initialised.
        self.sandbox  # pylint: disable=pointless-statement

    @property
    def is_initialised(self) -> bool:
        """Return whether the repository has been initialised."""
        return isinstance(self._sandbox, SandboxFolder)

    @property
    def sandbox(self):
        """Return the sandbox instance of this repository."""
        if self._sandbox is None:
            self._sandbox = SandboxFolder()

        return self._sandbox

    def erase(self):
        """Delete the repository itself and all its contents."""
        if getattr(self, '_sandbox', None) is not None:
            try:
                shutil.rmtree(self.sandbox.abspath)
            except FileNotFoundError:
                pass
            finally:
                self._sandbox = None

    def _put_object_from_filelike(self, handle: BinaryIO) -> str:
        """Store the byte contents of a file in the repository.

        :param handle: filelike object with the byte content to be stored.
        :return: the generated fully qualified identifier for the object within the repository.
        :raises TypeError: if the handle is not a byte stream.
        """
        key = str(uuid.uuid4())
        filepath = os.path.join(self.sandbox.abspath, key)

        with open(filepath, 'wb') as target:
            shutil.copyfileobj(handle, target)

        return key

    def has_objects(self, keys: List[str]) -> List[bool]:
        result = []
        dirlist = os.listdir(self.sandbox.abspath)
        for key in keys:
            result.append(key in dirlist)
        return result

    @contextlib.contextmanager
    def open(self, key: str) -> Iterator[BinaryIO]:
        super().open(key)

        with self.sandbox.open(key, mode='rb') as handle:
            yield handle

    def iter_object_streams(self, keys: List[str]) -> Iterator[Tuple[str, BinaryIO]]:
        for key in keys:
            with self.open(key) as handle:  # pylint: disable=not-context-manager
                yield key, handle

    def delete_objects(self, keys: List[str]) -> None:
        super().delete_objects(keys)
        for key in keys:
            os.remove(os.path.join(self.sandbox.abspath, key))

    def list_objects(self) -> Iterable[str]:
        return self.sandbox.get_content_list()

    def maintain(self, dry_run: bool = False, live: bool = True, **kwargs) -> None:
        raise NotImplementedError

    def get_info(self, detailed: bool = False, **kwargs) -> dict:
        raise NotImplementedError



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# yapf: disable
"""Module with pre-defined reusable commandline arguments that can be used as `click` decorators."""

# AUTO-GENERATED

# yapf: disable
# pylint: disable=wildcard-import

from .main import *
from .overridable import *

__all__ = (
    'CALCULATION',
    'CALCULATIONS',
    'CODE',
    'CODES',
    'COMPUTER',
    'COMPUTERS',
    'CONFIG_OPTION',
    'DATA',
    'DATUM',
    'GROUP',
    'GROUPS',
    'INPUT_FILE',
    'LABEL',
    'NODE',
    'NODES',
    'OUTPUT_FILE',
    'OverridableArgument',
    'PROCESS',
    'PROCESSES',
    'PROFILE',
    'PROFILES',
    'USER',
    'WORKFLOW',
    'WORKFLOWS',
)

# yapf: enable



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member,import-error,no-name-in-module,protected-access
"""This migration cleans the log records from non-Node entity records.

It removes from the DbLog table the legacy workflow records and records
that correspond to an unknown entity and places them to corresponding files.

This migration corresponds to the 0024_dblog_update Django migration (except without uuid addition).

Revision ID: 041a79fc615f
Revises: 7ca08c391c49
Create Date: 2018-12-28 15:53:14.596810
"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import text

from aiida.storage.psql_dos.migrations.utils.dblog_update import export_and_clean_workflow_logs

# revision identifiers, used by Alembic.
revision = '041a79fc615f'
down_revision = '7ca08c391c49'
branch_labels = None
depends_on = None


def upgrade():
    """
    Changing the log table columns to use uuid to reference remote objects and log entries.
    Upgrade function.
    """
    connection = op.get_bind()

    # Clean data
    export_and_clean_workflow_logs(connection, op.get_context().opts['aiida_profile'])

    # Remove objpk and objname from the metadata dictionary
    connection.execute(text("""UPDATE db_dblog SET metadata = metadata - 'objpk' - 'objname' """))

    # Create a new column, which is a foreign key to the dbnode table
    op.add_column('db_dblog', sa.Column('dbnode_id', sa.INTEGER(), autoincrement=False, nullable=True))
    # Transfer data to dbnode_id from objpk
    connection.execute(text("""UPDATE db_dblog SET dbnode_id=objpk"""))
    op.create_foreign_key(
        'db_dblog_dbnode_id_fkey',
        'db_dblog',
        'db_dbnode', ['dbnode_id'], ['id'],
        ondelete='CASCADE',
        initially='DEFERRED',
        deferrable=True
    )

    # Now that all the data have been migrated, make the column not nullable and not blank.
    # A log record should always correspond to a node record
    op.alter_column('db_dblog', 'dbnode_id', nullable=False)

    # Remove the objpk column
    op.drop_column('db_dblog', 'objpk')

    # Remove the objname column
    op.drop_column('db_dblog', 'objname')


def downgrade():
    """
    Downgrade function to the previous schema.
    """
    raise NotImplementedError('Downgrade of 041a79fc615f.')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Rename `db_dbcomputer.name` to `db_dbcomputer.label`

Revision ID: django_0048
Revises: django_0047

"""
from alembic import op

from aiida.storage.psql_dos.migrations.utils import ReflectMigrations

revision = 'django_0048'
down_revision = 'django_0047'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    reflect = ReflectMigrations(op)
    reflect.drop_unique_constraints('db_dbcomputer', ['name'])  # db_dbcomputer_name_key
    reflect.drop_indexes('db_dbcomputer', 'name')  # db_dbcomputer_name_f1800b1a_like
    op.alter_column('db_dbcomputer', 'name', new_column_name='label')
    op.create_unique_constraint('db_dbcomputer_label_bc480bab_uniq', 'db_dbcomputer', ['label'])
    op.create_index(
        'db_dbcomputer_label_bc480bab_like',
        'db_dbcomputer',
        ['label'],
        postgresql_using='btree',
        postgresql_ops={'label': 'varchar_pattern_ops'},
    )


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0048.')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Abstracts JSON usage to ensure compatibility with Python2 and Python3.

Use this module prefentially over standard json to ensure compatibility.

.. deprecated:: This module is deprecated in v2.0.0 and should no longer be used. Python 2 support has long since been
    dropped and for Python 3, one should simply use the ``json`` module of the standard library directly.

"""
import codecs
import json
import warnings

from aiida.common.warnings import AiidaDeprecationWarning

warnings.warn(
    'This module has been deprecated and should no longer be used. Use the `json` standard library instead.',
    AiidaDeprecationWarning
)


def dump(data, handle, **kwargs):
    """Serialize ``data`` as a JSON formatted stream to ``handle``.

    We use ``ensure_ascii=False`` to write unicode characters specifically as this improves the readability of the json
    and reduces the file size.
    """
    try:
        if 'b' in handle.mode:
            handle = codecs.getwriter('utf-8')(handle)
    except AttributeError:
        pass

    return json.dump(data, handle, ensure_ascii=False, **kwargs)


def dumps(data, **kwargs):
    """Serialize ``data`` as a JSON formatted string.

    We use ``ensure_ascii=False`` to write unicode characters specifically as this improves the readability of the json
    and reduces the file size.
    """
    return json.dumps(data, ensure_ascii=False, **kwargs)


def load(handle, **kwargs):
    """Deserialize ``handle`` text or binary file containing a JSON document to a Python object.

    :raises ValueError: if no valid JSON object could be decoded.
    """
    if 'b' in handle.mode:
        handle = codecs.getreader('utf-8')(handle)

    try:
        return json.load(handle, **kwargs)
    except json.JSONDecodeError as exc:
        raise ValueError from exc


def loads(string, **kwargs):
    """Deserialize text or binary ``string`` containing a JSON document to a Python object.

    :raises ValueError: if no valid JSON object could be decoded.
    """
    if isinstance(string, bytes):
        string = string.decode('utf-8')

    try:
        return json.loads(string, **kwargs)
    except json.JSONDecodeError as exc:
        raise ValueError from exc



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Add `db_dbnode.process_type`

Revision ID: django_0010
Revises: django_0009

"""
from alembic import op
import sqlalchemy as sa

revision = 'django_0010'
down_revision = 'django_0009'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.add_column('db_dbnode', sa.Column('process_type', sa.String(length=255), nullable=True))
    op.create_index('db_dbnode_process_type_df7298d0', 'db_dbnode', ['process_type'])
    op.create_index(
        'db_dbnode_process_type_df7298d0_like',
        'db_dbnode',
        ['process_type'],
        postgresql_using='btree',
        postgresql_ops={'process_type': 'varchar_pattern_ops'},
    )


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0010.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Simple wrapper around the alembic command line tool that first loads an AiiDA profile."""
import alembic
import click
from sqlalchemy.util.compat import nullcontext

from aiida.cmdline import is_verbose
from aiida.cmdline.params import options
from aiida.storage.psql_dos.migrator import PsqlDostoreMigrator


class AlembicRunner:
    """Wrapper around the alembic command line tool that first loads an AiiDA profile."""

    def __init__(self) -> None:
        self.profile = None

    def execute_alembic_command(self, command_name, connect=True, **kwargs):
        """Execute an Alembic CLI command.

        :param command_name: the sub command name
        :param kwargs: parameters to pass to the command
        """
        if self.profile is None:
            raise click.ClickException('No profile specified')
        migrator = PsqlDostoreMigrator(self.profile)

        context = migrator._alembic_connect() if connect else nullcontext(migrator._alembic_config())  # pylint: disable=protected-access
        with context as config:
            command = getattr(alembic.command, command_name)
            config.stdout = click.get_text_stream('stdout')
            command(config, **kwargs)


pass_runner = click.make_pass_decorator(AlembicRunner, ensure=True)


@click.group()
@options.PROFILE(required=True)
@pass_runner
def alembic_cli(runner, profile):
    """Simple wrapper around the alembic command line tool that first loads an AiiDA profile."""
    runner.profile = profile


@alembic_cli.command('revision')
@click.argument('message')
@pass_runner
def alembic_revision(runner, message):
    """Create a new database revision."""
    # to-do this does not currently work, because `alembic.RevisionContext._run_environment` has issues with heads
    # (it works if we comment out the initial autogenerate check)
    runner.execute_alembic_command('revision', message=message, autogenerate=True, head='main@head')


@alembic_cli.command('current')
@options.VERBOSITY()
@pass_runner
def alembic_current(runner):
    """Show the current revision."""
    runner.execute_alembic_command('current', verbose=is_verbose())


@alembic_cli.command('history')
@click.option('-r', '--rev-range')
@options.VERBOSITY()
@pass_runner
def alembic_history(runner, rev_range):
    """Show the history for the given revision range."""
    runner.execute_alembic_command('history', connect=False, rev_range=rev_range, verbose=is_verbose())


@alembic_cli.command('show')
@click.argument('revision', type=click.STRING)
@pass_runner
def alembic_show(runner, revision):
    """Show details of the given REVISION."""
    runner.execute_alembic_command('show', rev=revision)


@alembic_cli.command('upgrade')
@click.argument('revision', type=click.STRING)
@pass_runner
def alembic_upgrade(runner, revision):
    """Upgrade the database to the given REVISION."""
    runner.execute_alembic_command('upgrade', revision=revision)


@alembic_cli.command('downgrade')
@click.argument('revision', type=click.STRING)
@pass_runner
def alembic_downgrade(runner, revision):
    """Downgrade the database to the given REVISION."""
    runner.execute_alembic_command('downgrade', revision=revision)


if __name__ == '__main__':
    alembic_cli()  # pylint: disable=no-value-for-parameter



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""SQLA implementations for the Comment entity and collection."""
# pylint: disable=import-error,no-name-in-module

from datetime import datetime

from sqlalchemy.orm.exc import NoResultFound

from aiida.common import exceptions, lang
from aiida.orm.implementation.comments import BackendComment, BackendCommentCollection
from aiida.storage.psql_dos.models import comment as models

from . import entities, users, utils


class SqlaComment(entities.SqlaModelEntity[models.DbComment], BackendComment):
    """Comment implementation for Sqla."""

    MODEL_CLASS = models.DbComment

    # pylint: disable=too-many-arguments
    def __init__(self, backend, node, user, content=None, ctime=None, mtime=None):
        """Construct a SqlaComment.

        :param node: a Node instance
        :param user: a User instance
        :param content: the comment content
        :param ctime: The creation time as datetime object
        :param mtime: The modification time as datetime object
        """
        super().__init__(backend)
        lang.type_check(user, users.SqlaUser)  # pylint: disable=no-member

        arguments = {
            'dbnode': node.bare_model,
            'user': user.bare_model,
            'content': content,
        }

        if ctime:
            lang.type_check(ctime, datetime, f'the given ctime is of type {type(ctime)}')
            arguments['ctime'] = ctime

        if mtime:
            lang.type_check(mtime, datetime, f'the given mtime is of type {type(mtime)}')
            arguments['mtime'] = mtime

        self._model = utils.ModelWrapper(self.MODEL_CLASS(**arguments), backend)

    def store(self):
        """Can only store if both the node and user are stored as well."""
        if self.model.dbnode.id is None or self.model.user.id is None:
            self.model.dbnode = None
            raise exceptions.ModificationNotAllowed('The corresponding node and/or user are not stored')

        super().store()

    @property
    def uuid(self) -> str:
        return str(self.model.uuid)

    @property
    def ctime(self):
        return self.model.ctime

    @property
    def mtime(self):
        return self.model.mtime

    def set_mtime(self, value):
        self.model.mtime = value

    @property
    def node(self):
        return self.backend.nodes.ENTITY_CLASS.from_dbmodel(self.model.dbnode, self.backend)

    @property
    def user(self):
        return self.backend.users.ENTITY_CLASS.from_dbmodel(self.model.user, self.backend)

    def set_user(self, value):
        self.model.user = value

    @property
    def content(self):
        return self.model.content

    def set_content(self, value):
        self.model.content = value


class SqlaCommentCollection(BackendCommentCollection):
    """SqlAlchemy implementation for the CommentCollection."""

    ENTITY_CLASS = SqlaComment

    def create(self, node, user, content=None, **kwargs):
        """
        Create a Comment for a given node and user

        :param node: a Node instance
        :param user: a User instance
        :param content: the comment content
        :return: a Comment object associated to the given node and user
        """
        return self.ENTITY_CLASS(self.backend, node, user, content, **kwargs)

    def delete(self, comment_id):
        """
        Remove a Comment from the collection with the given id

        :param comment_id: the id of the comment to delete
        :type comment_id: int

        :raises TypeError: if ``comment_id`` is not an `int`
        :raises `~aiida.common.exceptions.NotExistent`: if Comment with ID ``comment_id`` is not found
        """
        if not isinstance(comment_id, int):
            raise TypeError('comment_id must be an int')

        session = self.backend.get_session()

        try:
            row = session.query(self.ENTITY_CLASS.MODEL_CLASS).filter_by(id=comment_id).one()
            session.delete(row)
            session.commit()
        except NoResultFound:
            session.rollback()
            raise exceptions.NotExistent(f"Comment with id '{comment_id}' not found")

    def delete_all(self):
        """
        Delete all Comment entries.

        :raises `~aiida.common.exceptions.IntegrityError`: if all Comments could not be deleted
        """
        session = self.backend.get_session()

        try:
            session.query(self.ENTITY_CLASS.MODEL_CLASS).delete()
            session.commit()
        except Exception as exc:
            session.rollback()
            raise exceptions.IntegrityError(f'Could not delete all Comments. Full exception: {exc}')

    def delete_many(self, filters):
        """
        Delete Comments based on ``filters``

        :param filters: similar to QueryBuilder filter
        :type filters: dict

        :return: (former) ``PK`` s of deleted Comments
        :rtype: list

        :raises TypeError: if ``filters`` is not a `dict`
        :raises `~aiida.common.exceptions.ValidationError`: if ``filters`` is empty
        """
        from aiida.orm import Comment, QueryBuilder

        # Checks
        if not isinstance(filters, dict):
            raise TypeError('filters must be a dictionary')
        if not filters:
            raise exceptions.ValidationError('filters must not be empty')

        # Apply filter and delete found entities
        builder = QueryBuilder(backend=self.backend).append(Comment, filters=filters, project='id')
        entities_to_delete = builder.all(flat=True)
        for entity in entities_to_delete:
            self.delete(entity)

        # Return list of deleted entities' (former) PKs for checking
        return entities_to_delete



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Testing infrastructure for easy testing of AiiDA plugins.
"""

# AUTO-GENERATED

# yapf: disable
# pylint: disable=wildcard-import

from .main import *

__all__ = (
    'ProfileManager',
    'TemporaryProfileManager',
    'TestManager',
    'TestManagerError',
    'get_test_backend_name',
    'get_test_profile_name',
    'get_user_dict',
    'test_manager',
)

# yapf: enable



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Convenience classes to help building the input dictionaries for Processes."""
from collections.abc import Mapping, MutableMapping
import json
from typing import TYPE_CHECKING, Any, Type
from uuid import uuid4

import yaml

from aiida.engine.processes.ports import PortNamespace
from aiida.orm import Dict, Node
from aiida.orm.nodes.data.base import BaseType

if TYPE_CHECKING:
    from aiida.engine.processes.process import Process

__all__ = ('ProcessBuilder', 'ProcessBuilderNamespace')


class PrettyEncoder(json.JSONEncoder):
    """JSON encoder for returning a pretty representation of an AiiDA ``ProcessBuilder``."""

    def default(self, o):  # pylint: disable=arguments-differ
        if isinstance(o, (ProcessBuilder, ProcessBuilderNamespace)):
            return dict(o)
        if isinstance(o, Dict):
            return o.get_dict()
        if isinstance(o, BaseType):
            return o.value
        if isinstance(o, Node):
            return o.get_description()


class ProcessBuilderNamespace(MutableMapping):
    """Input namespace for the `ProcessBuilder`.

    Dynamically generates the getters and setters for the input ports of a given PortNamespace
    """

    def __init__(self, port_namespace: PortNamespace) -> None:
        """Dynamically construct the get and set properties for the ports of the given port namespace.

        For each port in the given port namespace a get and set property will be constructed dynamically
        and added to the ProcessBuilderNamespace. The docstring for these properties will be defined
        by calling str() on the Port, which should return the description of the Port.

        :param port_namespace: the inputs PortNamespace for which to construct the builder

        """
        # pylint: disable=super-init-not-called
        self._port_namespace = port_namespace
        self._valid_fields = []
        self._data = {}

        dynamic_properties = {}

        # The name and port objects have to be passed to the defined functions as defaults for
        # their arguments, because this way the content at the time of defining the method is
        # saved. If they are used directly in the body, it will try to capture the value from
        # its enclosing scope at the time of being called.
        for name, port in port_namespace.items():

            self._valid_fields.append(name)

            if isinstance(port, PortNamespace):
                self._data[name] = ProcessBuilderNamespace(port)

                def fgetter(self, name=name):
                    return self._data.get(name)
            elif port.has_default():

                def fgetter(self, name=name, default=port.default):  # type: ignore # pylint: disable=cell-var-from-loop
                    return self._data.get(name, default)
            else:

                def fgetter(self, name=name):
                    return self._data.get(name, None)

            def fsetter(self, value, name=name):
                self._data[name] = value

            fgetter.__doc__ = str(port)
            getter = property(fgetter)
            getter.setter(fsetter)  # pylint: disable=too-many-function-args
            dynamic_properties[name] = getter

        # The dynamic property can only be attached to a class and not an instance, however, we cannot attach it to
        # the ``ProcessBuilderNamespace`` class since it would interfere with other instances that may already
        # exist. The workaround is to create a new class on the fly that derives from ``ProcessBuilderNamespace``
        # and add the dynamic property to that instead
        class_name = f'{self.__class__.__name__}-{uuid4()}'
        child_class = type(class_name, (self.__class__,), dynamic_properties)
        self.__class__ = child_class

    def __setattr__(self, attr: str, value: Any) -> None:
        """Assign the given value to the port with key `attr`.

        .. note:: Any attributes without a leading underscore being set correspond to inputs and should hence be
            validated with respect to the corresponding input port from the process spec

        """
        if attr.startswith('_'):
            object.__setattr__(self, attr, value)
        else:
            try:
                port = self._port_namespace[attr]
            except KeyError as exception:
                if not self._port_namespace.dynamic:
                    raise AttributeError(f'Unknown builder parameter: {attr}') from exception
                port = None
            else:
                value = port.serialize(value)  # type: ignore[union-attr]
                validation_error = port.validate(value)  # type: ignore[union-attr]
                if validation_error:
                    raise ValueError(f'invalid attribute value {validation_error.message}')

            # If the attribute that is being set corresponds to a port that is a ``PortNamespace`` we need to make sure
            # that the nested value remains a ``ProcessBuilderNamespace``. Otherwise, the nested namespaces will become
            # plain dictionaries and no longer have the properties of the ``ProcessBuilderNamespace`` that provide all
            # the autocompletion and validation when values are being set. Therefore we first construct a new instance
            # of a ``ProcessBuilderNamespace`` for the port of the attribute that is being set and than iteratively set
            # all the values within the mapping that is being assigned to the attribute.
            if isinstance(port, PortNamespace):
                self._data[attr] = ProcessBuilderNamespace(port)
                for sub_key, sub_value in value.items():
                    setattr(self._data[attr], sub_key, sub_value)
            else:
                self._data[attr] = value

    def __repr__(self):
        return self._data.__repr__()

    def __dir__(self):
        return sorted(set(self._valid_fields + [key for key, _ in self.__dict__.items() if key.startswith('_')]))

    def __iter__(self):
        for key in self._data:
            yield key

    def __len__(self):
        return len(self._data)

    def __getitem__(self, item):
        return self._data[item]

    def __setitem__(self, item, value):
        self.__setattr__(item, value)

    def __delitem__(self, item):
        self._data.__delitem__(item)

    def __delattr__(self, item):
        self._data.__delitem__(item)

    def _recursive_merge(self, dictionary, key, value):
        """Recursively merge the contents of ``dictionary`` setting its ``key`` to ``value``."""
        if isinstance(value, Mapping):
            for inner_key, inner_value in value.items():
                self._recursive_merge(dictionary[key], inner_key, inner_value)
        else:
            dictionary[key] = value

    def _merge(self, *args, **kwds):
        """Merge the content of a dictionary or keyword arguments in .

        .. note:: This method differs in behavior from ``_update`` in that ``_merge`` will recursively update the
            existing dictionary with the one that is specified in the arguments. The ``_update`` method will merge only
            the keys on the top level, but any lower lying nested namespace will be replaced entirely.

        The method is prefixed with an underscore in order to not reserve the name for a potential port.

        :param args: a single mapping that should be mapped on the namespace.
        :param kwds: keyword value pairs that should be mapped onto the ports.
        """
        if len(args) > 1:
            raise TypeError(f'update expected at most 1 arguments, got {int(len(args))}')

        if args:
            for key, value in args[0].items():
                self._recursive_merge(self, key, value)

        for key, value in kwds.items():
            self._recursive_merge(self, key, value)

    def _update(self, *args, **kwds):
        """Update the values of the builder namespace passing a mapping as argument or individual keyword value pairs.

        The method functions just as `collections.abc.MutableMapping.update` and is merely prefixed with an underscore
        in order to not reserve the name for a potential port.

        :param args: a single mapping that should be mapped on the namespace.
        :param kwds: keyword value pairs that should be mapped onto the ports.
        """
        if args:
            for key, value in args[0].items():
                if isinstance(value, Mapping):
                    self[key].update(value)
                else:
                    self.__setattr__(key, value)

        for key, value in kwds.items():
            if isinstance(value, Mapping):
                self[key].update(value)
            else:
                self.__setattr__(key, value)

    def _inputs(self, prune: bool = False) -> dict:
        """Return the entire mapping of inputs specified for this builder.

        :param prune: boolean, when True, will prune nested namespaces that contain no actual values whatsoever
        :return: mapping of inputs ports and their input values.
        """
        if prune:
            return self._prune(dict(self))

        return dict(self)

    def _prune(self, value):
        """Prune a nested mapping from all mappings that are completely empty.

        .. note:: a nested mapping that is completely empty means it contains at most other empty mappings. Other null
            values, such as `None` or empty lists, should not be pruned.

        :param value: a nested mapping of port values
        :return: the same mapping but without any nested namespace that is completely empty.
        """
        if isinstance(value, Mapping) and not isinstance(value, Node):
            result = {}
            for key, sub_value in value.items():
                pruned = self._prune(sub_value)
                # If `pruned` is an "empty'ish" mapping and not an instance of `Node`, skip it, otherwise keep it.
                if not (isinstance(pruned, Mapping) and not pruned and not isinstance(pruned, Node)):
                    result[key] = pruned
            return result

        return value


class ProcessBuilder(ProcessBuilderNamespace):  # pylint: disable=too-many-ancestors
    """A process builder that helps setting up the inputs for creating a new process."""

    def __init__(self, process_class: Type['Process']):
        """Construct a `ProcessBuilder` instance for the given `Process` class.

        :param process_class: the `Process` subclass
        """
        self._process_class = process_class
        self._process_spec = self._process_class.spec()
        super().__init__(self._process_spec.inputs)

    @property
    def process_class(self) -> Type['Process']:
        """Return the process class for which this builder is constructed."""
        return self._process_class

    def _repr_pretty_(self, p, _) -> str:  # pylint: disable=invalid-name
        """Pretty representation for in the IPython console and notebooks."""
        return p.text(
            f'Process class: {self._process_class.__name__}\n'
            f'Inputs:\n{yaml.safe_dump(json.JSONDecoder().decode(PrettyEncoder().encode(self)))}'
        )



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Bring schema inline with psql_dos main_0001

Revision ID: main_0001
Revises:
Create Date: 2021-02-02

"""
revision = 'main_0001'
down_revision = 'main_0000b'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of main_0001.')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Move `db_dbattribute`/`db_dbextra` to `db_dbnode.attributes`/`db_dbnode.extras`, and add `dbsetting.val`

Revision ID: django_0037
Revises: django_0036

"""
import math

from alembic import op
import sqlalchemy as sa
from sqlalchemy import cast, func, select
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql import column, table

from aiida.cmdline.utils import echo
from aiida.common.progress_reporter import get_progress_reporter
from aiida.common.timezone import datetime_to_isoformat
from aiida.storage.psql_dos.migrations.utils import ReflectMigrations

revision = 'django_0037'
down_revision = 'django_0036'
branch_labels = None
depends_on = None

node_tbl = table(
    'db_dbnode',
    column('id'),
    column('attributes', postgresql.JSONB(astext_type=sa.Text())),
    column('extras', postgresql.JSONB(astext_type=sa.Text())),
)

attr_tbl = table(
    'db_dbattribute',
    column('id'),
    column('dbnode_id'),
    column('key'),
    column('datatype'),
    column('tval'),
    column('ival'),
    column('fval'),
    column('dval'),
    column('bval'),
)

extra_tbl = table(
    'db_dbextra',
    column('id'),
    column('dbnode_id'),
    column('key'),
    column('datatype'),
    column('tval'),
    column('ival'),
    column('fval'),
    column('dval'),
    column('bval'),
)

setting_tbl = table(
    'db_dbsetting',
    column('id'),
    column('description'),
    column('time'),
    column('key'),
    column('datatype'),
    column('tval'),
    column('ival'),
    column('fval'),
    column('dval'),
    column('bval'),
    column('val'),
)


def upgrade():
    """Migrations for the upgrade."""
    conn = op.get_bind()

    op.add_column('db_dbnode', sa.Column('attributes', postgresql.JSONB(astext_type=sa.Text()), nullable=True))
    op.add_column('db_dbnode', sa.Column('extras', postgresql.JSONB(astext_type=sa.Text()), nullable=True))

    # transition attributes and extras to node
    node_count = conn.execute(select(func.count()).select_from(node_tbl)).scalar()
    if node_count:
        with get_progress_reporter()(total=node_count, desc='Updating attributes and extras') as progress:
            for node in conn.execute(select(node_tbl)).all():
                attr_list = conn.execute(select(attr_tbl).where(attr_tbl.c.dbnode_id == node.id)).all()
                attributes, _ = attributes_to_dict(sorted(attr_list, key=lambda a: a.key))
                extra_list = conn.execute(select(extra_tbl).where(extra_tbl.c.dbnode_id == node.id)).all()
                extras, _ = attributes_to_dict(sorted(extra_list, key=lambda a: a.key))
                conn.execute(
                    node_tbl.update().where(node_tbl.c.id == node.id).values(attributes=attributes, extras=extras)
                )
                progress.update()

    op.drop_table('db_dbattribute')
    op.drop_table('db_dbextra')

    op.add_column('db_dbsetting', sa.Column('val', postgresql.JSONB(astext_type=sa.Text()), nullable=True))

    # transition settings
    setting_count = conn.execute(select(func.count()).select_from(setting_tbl)).scalar()
    if setting_count:
        with get_progress_reporter()(total=setting_count, desc='Updating settings') as progress:
            for setting in conn.execute(select(setting_tbl)).all():
                dt = setting.datatype
                val = None
                if dt == 'txt':
                    val = setting.tval
                elif dt == 'float':
                    val = setting.fval
                    if math.isnan(val) or math.isinf(val):
                        val = str(val)
                elif dt == 'int':
                    val = setting.ival
                elif dt == 'bool':
                    val = setting.bval
                elif dt == 'date':
                    val = datetime_to_isoformat(setting.dval)
                conn.execute(
                    setting_tbl.update().where(setting_tbl.c.id == setting.id
                                               ).values(val=cast(val, postgresql.JSONB(astext_type=sa.Text())))
                )
                progress.update()

    op.drop_column('db_dbsetting', 'tval')
    op.drop_column('db_dbsetting', 'fval')
    op.drop_column('db_dbsetting', 'ival')
    op.drop_column('db_dbsetting', 'bval')
    op.drop_column('db_dbsetting', 'dval')
    op.drop_column('db_dbsetting', 'datatype')

    ReflectMigrations(op).drop_indexes('db_dbsetting', 'key')  # db_dbsetting_key_1b84beb4
    op.create_index(
        'db_dbsetting_key_1b84beb4_like',
        'db_dbsetting',
        ['key'],
        postgresql_using='btree',
        postgresql_ops={'key': 'varchar_pattern_ops'},
    )


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0037.')


def attributes_to_dict(attr_list: list):
    """
    Transform the attributes of a node into a dictionary. It assumes the key
    are ordered alphabetically, and that they all belong to the same node.
    """
    d = {}

    error = False
    for a in attr_list:
        try:
            tmp_d = select_from_key(a.key, d)
        except ValueError:
            echo.echo_error(f"Couldn't transfer attribute {a.id} with key {a.key} for dbnode {a.dbnode_id}")
            error = True
            continue
        key = a.key.split('.')[-1]

        if isinstance(tmp_d, (list, tuple)):
            key = int(key)

        dt = a.datatype

        if dt == 'dict':
            tmp_d[key] = {}
        elif dt == 'list':
            tmp_d[key] = [None] * a.ival
        else:
            val = None
            if dt == 'txt':
                val = a.tval
            elif dt == 'float':
                val = a.fval
                if math.isnan(val) or math.isinf(val):
                    val = str(val)
            elif dt == 'int':
                val = a.ival
            elif dt == 'bool':
                val = a.bval
            elif dt == 'date':
                val = datetime_to_isoformat(a.dval)

            tmp_d[key] = val

    return d, error


def select_from_key(key, d):
    """
    Return element of the dict to do the insertion on. If it is foo.1.bar, it
    will return d["foo"][1]. If it is only foo, it will return d directly.
    """
    path = key.split('.')[:-1]

    tmp_d = d
    for p in path:
        if isinstance(tmp_d, (list, tuple)):
            tmp_d = tmp_d[int(p)]
        else:
            tmp_d = tmp_d[p]

    return tmp_d



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Update all link labels with the value `_return`

This is the legacy default single link label.
The old process functions used to use `_return` as the default link label, however, since labels that start or end with
and underscore are illegal because they are used for namespacing.

This is identical to migration 118349c10896

Revision ID: django_0043
Revises: django_0042

"""
from alembic import op
import sqlalchemy as sa

revision = 'django_0043'
down_revision = 'django_0042'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    conn = op.get_bind()
    statement = sa.text("""
        UPDATE db_dblink SET label='result' WHERE label = '_return';
    """)
    conn.execute(statement)


def downgrade():
    """Migrations for the downgrade."""
    statement = sa.text("""
        UPDATE db_dblink SET label='_result' WHERE label = 'return';
    """)
    op.get_bind().execute(statement)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Serialisation functions for AiiDA types

WARNING: Changing the representation of things here may break people's current saved e.g. things like
checkpoints and messages in the RabbitMQ queue so do so with caution.  It is fine to add representers
for new types though.
"""
from enum import Enum
from functools import partial

from plumpy import Bundle, get_object_loader
from plumpy.utils import AttributesFrozendict
import yaml

from aiida import orm
from aiida.common import AttributeDict

_ENUM_TAG = '!enum'
_NODE_TAG = '!aiida_node'
_GROUP_TAG = '!aiida_group'
_COMPUTER_TAG = '!aiida_computer'
_ATTRIBUTE_DICT_TAG = '!aiida_attributedict'
_PLUMPY_ATTRIBUTES_FROZENDICT_TAG = '!plumpy:attributes_frozendict'
_PLUMPY_BUNDLE = '!plumpy:bundle'


def represent_enum(dumper, enum):
    """Represent an arbitrary enum in yaml.

    :param dumper: the dumper to use.
    :type dumper: :class:`yaml.dumper.Dumper`
    :param bundle: the bundle to represent
    :return: the representation
    """
    loader = get_object_loader()
    return dumper.represent_scalar(_ENUM_TAG, f'{loader.identify_object(enum)}|{enum.value}')


def enum_constructor(loader, serialized):
    """Construct an enum from the serialized representation.

    :param loader: the yaml loader.
    :type loader: :class:`yaml.loader.Loader`
    :param bundle: the enum representation.
    :return: the enum.
    """
    deserialized = loader.construct_scalar(serialized)
    identifier, value = deserialized.split('|')
    cls = get_object_loader().load_object(identifier)
    enum = cls(value)
    return enum


def represent_node(dumper, node):
    """Represent a node in yaml.

    :param dumper: the dumper to use
    :param node: the node to represent
    :type node: :class:`aiida.orm.nodes.node.Node`
    :return: the representation
    """
    if not node.is_stored:
        raise ValueError(f'node {type(node)}<{node.uuid}> cannot be represented because it is not stored')
    return dumper.represent_scalar(_NODE_TAG, f'{node.uuid}')


def node_constructor(loader, node):
    """Load a node from the yaml representation.

    :param loader: the yaml loader
    :param node: the yaml representation
    :return: the aiida node
    :rtype: :class:`aiida.orm.nodes.node.Node`
    """
    yaml_node = loader.construct_scalar(node)
    return orm.load_node(uuid=yaml_node)


def represent_group(dumper, group):
    """Represent a group in yaml.

    :param dumper: the dumper to use
    :param group: the group to represent
    :type group: :class:`aiida.orm.Group`
    :return: the representation
    """
    if not group.is_stored:
        raise ValueError(f'group {group} cannot be represented because it is not stored')
    return dumper.represent_scalar(_GROUP_TAG, f'{group.uuid}')


def group_constructor(loader, group):
    """Load a group from the yaml representation.

    :param loader: the yaml loader
    :param group: the yaml representation
    :return: the aiida group
    :rtype: :class:`aiida.orm.Group`
    """
    yaml_node = loader.construct_scalar(group)
    return orm.load_group(uuid=yaml_node)


def represent_computer(dumper, computer):
    """Represent a computer in yaml.

    :param dumper: the dumper to use
    :param computer: the computer to represent
    :type computer: :class:`aiida.orm.Computer`
    :return: the representation
    """
    if not computer.is_stored:
        raise ValueError(f'computer {computer} cannot be represented because it is not stored')
    return dumper.represent_scalar(_COMPUTER_TAG, f'{computer.uuid}')


def computer_constructor(loader, computer):
    """Load a computer from the yaml representation.

    :param loader: the yaml loader
    :param computer: the yaml representation
    :return: the aiida computer
    :rtype: :class:`aiida.orm.Computer`
    """
    yaml_node = loader.construct_scalar(computer)
    return orm.Computer.get(uuid=yaml_node)


def represent_mapping(tag, dumper, mapping):
    """Represent a mapping in yaml.

    :param tag: the yaml tag to use
    :param dumper: the dumper to use
    :type dumper: :class:`yaml.dumper.Dumper`
    :param mapping: the mapping to represent
    :return: the representation
    """
    return dumper.represent_mapping(tag, mapping)


def mapping_constructor(mapping_type, loader, mapping):
    """Construct a mapping from the representation.

    :param mapping_type: the class of the mapping to construct, must accept a dictionary as a sole constructor argument
        to be compatible.
    :param loader: the yaml loader
    :type loader: :class:`yaml.loader.Loader`
    :param mapping: the mapping representation
    :return: the reconstructed mapping
    """
    yaml_node = loader.construct_mapping(mapping, deep=True)
    return mapping_type(yaml_node)


def represent_bundle(dumper, bundle):
    """Represent an `plumpy.Bundle` in yaml

    :param tag: the yaml  tag to use
    :param dumper: the dumper to use
    :type dumper: :class:`yaml.dumper.Dumper`
    :param bundle: the bundle to represent
    :return: the representation
    """
    as_dict = dict(bundle)
    return dumper.represent_mapping(_PLUMPY_BUNDLE, as_dict)


def bundle_constructor(loader, bundle):
    """Construct an `plumpy.Bundle` from the representation

    :param loader: the yaml loader
    :type loader: :class:`yaml.loader.Loader`
    :param bundle: the bundle representation
    :return: the mapping type
    """
    yaml_node = loader.construct_mapping(bundle)
    bundle = Bundle.__new__(Bundle)
    bundle.update(yaml_node)
    return bundle


class AiiDADumper(yaml.Dumper):
    """Custom AiiDA yaml dumper.

    Needed so that we don't have to encode each type in the AiiDA graph hierarchy separately using a custom representer.
    """

    def represent_data(self, data):
        if isinstance(data, orm.Node):
            return represent_node(self, data)
        if isinstance(data, orm.Computer):
            return represent_computer(self, data)
        if isinstance(data, orm.Group):
            return represent_group(self, data)

        return super().represent_data(data)


class AiiDALoader(yaml.Loader):
    """AiiDA specific yaml loader

    .. note:: The `AiiDALoader` should only be used on trusted input, since it uses the `yaml.Loader` which is not safe.
        When importing a shared database, we strip all process node checkpoints to avoid this being a security risk.
    """


yaml.add_representer(Enum, represent_enum, Dumper=AiiDADumper)
yaml.add_representer(Bundle, represent_bundle, Dumper=AiiDADumper)
yaml.add_representer(AttributeDict, partial(represent_mapping, _ATTRIBUTE_DICT_TAG), Dumper=AiiDADumper)
yaml.add_constructor(_ATTRIBUTE_DICT_TAG, partial(mapping_constructor, AttributeDict), Loader=AiiDALoader)
yaml.add_representer(
    AttributesFrozendict, partial(represent_mapping, _PLUMPY_ATTRIBUTES_FROZENDICT_TAG), Dumper=AiiDADumper
)
yaml.add_constructor(
    _PLUMPY_ATTRIBUTES_FROZENDICT_TAG, partial(mapping_constructor, AttributesFrozendict), Loader=AiiDALoader
)
yaml.add_constructor(_PLUMPY_BUNDLE, bundle_constructor, Loader=AiiDALoader)
yaml.add_constructor(_NODE_TAG, node_constructor, Loader=AiiDALoader)
yaml.add_constructor(_GROUP_TAG, group_constructor, Loader=AiiDALoader)
yaml.add_constructor(_COMPUTER_TAG, computer_constructor, Loader=AiiDALoader)
yaml.add_constructor(_ENUM_TAG, enum_constructor, Loader=AiiDALoader)


def serialize(data, encoding=None):
    """Serialize the given data structure into a yaml dump.

    The function supports standard data containers such as maps and lists as well as AiiDA nodes which will be
    serialized into strings, before the whole data structure is dumped into a string using yaml.

    :param data: the general data to serialize
    :param encoding: optional encoding for the serialized string
    :return: string representation of the serialized data structure or byte array if specific encoding is specified
    """
    if encoding is not None:
        serialized = yaml.dump(data, encoding=encoding, Dumper=AiiDADumper)
    else:
        serialized = yaml.dump(data, Dumper=AiiDADumper)

    return serialized


def deserialize_unsafe(serialized):
    """Deserialize a yaml dump that represents a serialized data structure.

    .. note:: This function should not be used on untrusted input, since it is built upon `yaml.Loader` which is unsafe.

    :param serialized: a yaml serialized string representation
    :return: the deserialized data structure
    """
    return yaml.load(serialized, Loader=AiiDALoader)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module to define commonly used data structures."""
from enum import Enum, IntEnum

from .extendeddicts import DefaultFieldsAttributeDict

__all__ = ('StashMode', 'CalcJobState', 'CalcInfo', 'CodeInfo', 'CodeRunMode')


class StashMode(Enum):
    """Mode to use when stashing files from the working directory of a completed calculation job for safekeeping."""

    COPY = 'copy'


class CalcJobState(Enum):
    """The sub state of a CalcJobNode while its Process is in an active state (i.e. Running or Waiting)."""

    UPLOADING = 'uploading'
    SUBMITTING = 'submitting'
    WITHSCHEDULER = 'withscheduler'
    STASHING = 'stashing'
    RETRIEVING = 'retrieving'
    PARSING = 'parsing'


class CalcInfo(DefaultFieldsAttributeDict):
    """
    This object will store the data returned by the calculation plugin and to be
    passed to the ExecManager.

    In the following descriptions all paths have to be considered relative

    * retrieve_list: a list of strings or tuples that indicate files that are to be retrieved from the remote after the
        calculation has finished and stored in the ``retrieved_folder`` output node of type ``FolderData``. If the entry
        in the list is just a string, it is assumed to be the filepath on the remote and it will be copied to the base
        directory of the retrieved folder, where the name corresponds to the basename of the remote relative path. This
        means that any remote folder hierarchy is ignored entirely.

        Remote folder hierarchy can be (partially) maintained by using a tuple instead, with the following format

            (source, target, depth)

        The ``source`` and ``target`` elements are relative filepaths in the remote and retrieved folder. The contents
        of ``source`` (whether it is a file or folder) are copied in its entirety to the ``target`` subdirectory in the
        retrieved folder. If no subdirectory should be created, ``'.'`` should be specified for ``target``.

        The ``source`` filepaths support glob patterns ``*`` in case the exact name of the files that are to be
        retrieved are not know a priori.

        The ``depth`` element can be used to control what level of nesting of the source folder hierarchy should be
        maintained. If ``depth`` equals ``0`` or ``1`` (they are equivalent), only the basename of the ``source``
        filepath is kept. For each additional level, another subdirectory of the remote hierarchy is kept. For example:

            ('path/sub/file.txt', '.', 2)

        will retrieve the ``file.txt`` and store it under the path:

            sub/file.txt

    * retrieve_temporary_list: a list of strings or tuples that indicate files that will be retrieved
        and stored temporarily in a FolderData, that will be available only during the parsing call.
        The format of the list is the same as that of 'retrieve_list'

    * local_copy_list: a list of tuples with format ('node_uuid', 'filename', relativedestpath')
    * remote_copy_list: a list of tuples with format ('remotemachinename', 'remoteabspath', 'relativedestpath')
    * remote_symlink_list: a list of tuples with format ('remotemachinename', 'remoteabspath', 'relativedestpath')
    * provenance_exclude_list: a sequence of relative paths of files in the sandbox folder of a `CalcJob` instance that
        should not be stored permanantly in the repository folder of the corresponding `CalcJobNode` that will be
        created, but should only be copied to the remote working directory on the target computer. This is useful for
        input files that should be copied to the working directory but should not be copied as well to the repository
        either, for example, because they contain proprietary information or because they are big and their content is
        already indirectly present in the repository through one of the data nodes passed as input to the calculation.
    * codes_info: a list of dictionaries used to pass the info of the execution of a code
    * codes_run_mode: the mode of execution in which the codes will be run (`CodeRunMode.SERIAL` by default,
        but can also be `CodeRunMode.PARALLEL`)
    * skip_submit: a flag that, when set to True, orders the engine to skip the submit/update steps (so no code will
        run, it will only upload the files and then retrieve/parse).
    """

    _default_fields = (
        'job_environment', 'email', 'email_on_started', 'email_on_terminated', 'uuid', 'prepend_text', 'append_text',
        'num_machines', 'num_mpiprocs_per_machine', 'priority', 'max_wallclock_seconds', 'max_memory_kb', 'rerunnable',
        'retrieve_list', 'retrieve_temporary_list', 'local_copy_list', 'remote_copy_list', 'remote_symlink_list',
        'provenance_exclude_list', 'codes_info', 'codes_run_mode', 'skip_submit'
    )


class CodeInfo(DefaultFieldsAttributeDict):
    """
    This attribute-dictionary contains the information needed to execute a code.
    Possible attributes are:

    * ``cmdline_params``: a list of strings, containing parameters to be written on
      the command line right after the call to the code, as for example::

        code.x cmdline_params[0] cmdline_params[1] ... < stdin > stdout

    * ``stdin_name``: (optional) the name of the standard input file. Note, it is
      only possible to use the stdin with the syntax::

        code.x < stdin_name

      If no stdin_name is specified, the string "< stdin_name" will not be
      passed to the code.
      Note: it is not possible to substitute/remove the '<' if stdin_name is specified;
      if that is needed, avoid stdin_name and use instead the cmdline_params to
      specify a suitable syntax.
    * ``stdout_name``: (optional) the name of the standard output file. Note, it is
      only possible to pass output to stdout_name with the syntax::

        code.x ... > stdout_name

      If no stdout_name is specified, the string "> stdout_name" will not be
      passed to the code.
      Note: it is not possible to substitute/remove the '>' if stdout_name is specified;
      if that is needed, avoid stdout_name and use instead the cmdline_params to
      specify a suitable syntax.
    * ``stderr_name``: (optional) a string, the name of the error file of the code.
    * ``join_files``: (optional) if True, redirects the error to the output file.
      If join_files=True, the code will be called as::

        code.x ... > stdout_name 2>&1

      otherwise, if join_files=False and stderr is passed::

        code.x ... > stdout_name 2> stderr_name

    * ``withmpi``: if True, executes the code with mpirun (or another MPI installed
      on the remote computer)
    * ``code_uuid``: the uuid of the code associated to the CodeInfo
    """
    _default_fields = (
        'cmdline_params',  # as a list of strings
        'stdin_name',
        'stdout_name',
        'stderr_name',
        'join_files',
        'withmpi',
        'code_uuid'
    )


class CodeRunMode(IntEnum):
    """Enum to indicate the way the codes of a calculation should be run.

    For PARALLEL, the codes for a given calculation will be run in parallel by running them in the background::

        code1.x &
        code2.x &

    For the SERIAL option, codes will be executed sequentially by running for example the following::

        code1.x
        code2.x
    """

    SERIAL = 0
    PARALLEL = 1



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Remove the `node.` prefix from `db_dbnode.type`

Note, this is identical to the sqlalchemy migration 61fc0913fae9.

Revision ID: django_0028
Revises: django_0027

"""
from alembic import op
import sqlalchemy as sa

revision = 'django_0028'
down_revision = 'django_0027'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    conn = op.get_bind()

    # The `node.` prefix is being dropped from the node type string
    statement = sa.text(
        r"""
        UPDATE db_dbnode
        SET type = regexp_replace(type, '^node.data.', 'data.')
        WHERE type LIKE 'node.data.%';

        UPDATE db_dbnode
        SET type = regexp_replace(type, '^node.process.', 'process.')
        WHERE type LIKE 'node.process.%';
    """
    )
    conn.execute(statement)


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0028.')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Generic backend related objects"""
import abc
from typing import TYPE_CHECKING, Any, ContextManager, List, Optional, Sequence, TypeVar, Union

if TYPE_CHECKING:
    from aiida.manage.configuration.profile import Profile
    from aiida.orm.autogroup import AutogroupManager
    from aiida.orm.entities import EntityTypes
    from aiida.orm.implementation import (
        BackendAuthInfoCollection,
        BackendCommentCollection,
        BackendComputerCollection,
        BackendGroupCollection,
        BackendLogCollection,
        BackendNodeCollection,
        BackendQueryBuilder,
        BackendUserCollection,
    )
    from aiida.repository.backend.abstract import AbstractRepositoryBackend

__all__ = ('StorageBackend',)

TransactionType = TypeVar('TransactionType')


class StorageBackend(abc.ABC):  # pylint: disable=too-many-public-methods
    """Abstraction for a backend to read/write persistent data for a profile's provenance graph.

    AiiDA splits data storage into two sources:

    - Searchable data, which is stored in the database and can be queried using the QueryBuilder
    - Non-searchable (binary) data, which is stored in the repository and can be loaded using the RepositoryBackend

    The two sources are inter-linked by the ``Node.repository_metadata``.
    Once stored, the leaf values of this dictionary must be valid pointers to object keys in the repository.

    The class methods,`version_profile` and `migrate`,
    should be able to be called for existing storage, at any supported schema version (or empty storage).
    But an instance of this class should be created only for the latest schema version.

    """

    @classmethod
    @abc.abstractmethod
    def version_head(cls) -> str:
        """Return the head schema version of this storage backend type."""

    @classmethod
    @abc.abstractmethod
    def version_profile(cls, profile: 'Profile') -> Optional[str]:
        """Return the schema version of the given profile's storage, or None for empty/uninitialised storage.

        :raises: `~aiida.common.exceptions.UnreachableStorage` if the storage cannot be accessed
        """

    @classmethod
    @abc.abstractmethod
    def migrate(cls, profile: 'Profile') -> None:
        """Migrate the storage of a profile to the latest schema version.

        If the schema version is already the latest version, this method does nothing.
        If the storage is empty/uninitialised, then it will be initialised at head.

        :raises: `~aiida.common.exceptions.UnreachableStorage` if the storage cannot be accessed
        """

    @abc.abstractmethod
    def __init__(self, profile: 'Profile') -> None:
        """Initialize the backend, for this profile.

        :raises: `~aiida.common.exceptions.UnreachableStorage` if the storage cannot be accessed
        :raises: `~aiida.common.exceptions.IncompatibleStorageSchema`
            if the profile's storage schema is not at the latest version (and thus should be migrated)
        :raises: :raises: :class:`aiida.common.exceptions.CorruptStorage` if the storage is internally inconsistent
        """
        from aiida.orm.autogroup import AutogroupManager
        self._profile = profile
        self._autogroup = AutogroupManager(self)

    @abc.abstractmethod
    def __str__(self) -> str:
        """Return a string showing connection details for this instance."""

    @property
    def profile(self) -> 'Profile':
        """Return the profile for this backend."""
        return self._profile

    @property
    def autogroup(self) -> 'AutogroupManager':
        """Return the autogroup manager for this backend."""
        return self._autogroup

    def version(self) -> str:
        """Return the schema version of the profile's storage."""
        version = self.version_profile(self.profile)
        assert version is not None
        return version

    @abc.abstractmethod
    def close(self):
        """Close the storage access."""

    @property
    @abc.abstractmethod
    def is_closed(self) -> bool:
        """Return whether the storage is closed."""

    @abc.abstractmethod
    def _clear(self, recreate_user: bool = True) -> None:
        """Clear the storage, removing all data.

        .. warning:: This is a destructive operation, and should only be used for testing purposes.

        :param recreate_user: Re-create the default `User` for the profile, after clearing the storage.
        """
        from aiida.orm.autogroup import AutogroupManager
        self._autogroup = AutogroupManager(self)

    @property
    @abc.abstractmethod
    def authinfos(self) -> 'BackendAuthInfoCollection':
        """Return the collection of authorisation information objects"""

    @property
    @abc.abstractmethod
    def comments(self) -> 'BackendCommentCollection':
        """Return the collection of comments"""

    @property
    @abc.abstractmethod
    def computers(self) -> 'BackendComputerCollection':
        """Return the collection of computers"""

    @property
    @abc.abstractmethod
    def groups(self) -> 'BackendGroupCollection':
        """Return the collection of groups"""

    @property
    @abc.abstractmethod
    def logs(self) -> 'BackendLogCollection':
        """Return the collection of logs"""

    @property
    @abc.abstractmethod
    def nodes(self) -> 'BackendNodeCollection':
        """Return the collection of nodes"""

    @property
    @abc.abstractmethod
    def users(self) -> 'BackendUserCollection':
        """Return the collection of users"""

    @abc.abstractmethod
    def query(self) -> 'BackendQueryBuilder':
        """Return an instance of a query builder implementation for this backend"""

    @abc.abstractmethod
    def transaction(self) -> ContextManager[Any]:
        """
        Get a context manager that can be used as a transaction context for a series of backend operations.
        If there is an exception within the context then the changes will be rolled back and the state will
        be as before entering.  Transactions can be nested.

        :return: a context manager to group database operations
        """

    @property
    @abc.abstractmethod
    def in_transaction(self) -> bool:
        """Return whether a transaction is currently active."""

    @abc.abstractmethod
    def bulk_insert(self, entity_type: 'EntityTypes', rows: List[dict], allow_defaults: bool = False) -> List[int]:
        """Insert a list of entities into the database, directly into a backend transaction.

        :param entity_type: The type of the entity
        :param data: A list of dictionaries, containing all fields of the backend model,
            except the `id` field (a.k.a primary key), which will be generated dynamically
        :param allow_defaults: If ``False``, assert that each row contains all fields (except primary key(s)),
            otherwise, allow default values for missing fields.

        :raises: ``IntegrityError`` if the keys in a row are not a subset of the columns in the table

        :returns: The list of generated primary keys for the entities
        """

    @abc.abstractmethod
    def bulk_update(self, entity_type: 'EntityTypes', rows: List[dict]) -> None:
        """Update a list of entities in the database, directly with a backend transaction.

        :param entity_type: The type of the entity
        :param data: A list of dictionaries, containing fields of the backend model to update,
            and the `id` field (a.k.a primary key)

        :raises: ``IntegrityError`` if the keys in a row are not a subset of the columns in the table
        """

    @abc.abstractmethod
    def delete_nodes_and_connections(self, pks_to_delete: Sequence[int]):
        """Delete all nodes corresponding to pks in the input and any links to/from them.

        This method is intended to be used within a transaction context.

        :param pks_to_delete: a sequence of node pks to delete

        :raises: ``AssertionError`` if a transaction is not active
        """

    @abc.abstractmethod
    def get_repository(self) -> 'AbstractRepositoryBackend':
        """Return the object repository configured for this backend."""

    @abc.abstractmethod
    def set_global_variable(
        self, key: str, value: Union[None, str, int, float], description: Optional[str] = None, overwrite=True
    ) -> None:
        """Set a global variable in the storage.

        :param key: the key of the setting
        :param value: the value of the setting
        :param description: the description of the setting (optional)
        :param overwrite: if True, overwrite the setting if it already exists

        :raises: `ValueError` if the key already exists and `overwrite` is False
        """

    @abc.abstractmethod
    def get_global_variable(self, key: str) -> Union[None, str, int, float]:
        """Return a global variable from the storage.

        :param key: the key of the setting

        :raises: `KeyError` if the setting does not exist
        """

    @abc.abstractmethod
    def maintain(self, full: bool = False, dry_run: bool = False, **kwargs) -> None:
        """Perform maintenance tasks on the storage.

        If `full == True`, then this method may attempt to block the profile associated with the
        storage to guarantee the safety of its procedures. This will not only prevent any other
        subsequent process from accessing that profile, but will also first check if there is
        already any process using it and raise if that is the case. The user will have to manually
        stop any processes that is currently accessing the profile themselves or wait for it to
        finish on its own.

        :param full: flag to perform operations that require to stop using the profile to be maintained.
        :param dry_run: flag to only print the actions that would be taken without actually executing them.
        """

    def get_info(self, detailed: bool = False) -> dict:
        """Return general information on the storage.

        :param detailed: flag to request more detailed information about the content of the storage.
        :returns: a nested dict with the relevant information.
        """
        return {'entities': self.get_orm_entities(detailed=detailed)}

    def get_orm_entities(self, detailed: bool = False) -> dict:
        """Return a mapping with an overview of the storage contents regarding ORM entities.

        :param detailed: flag to request more detailed information about the content of the storage.
        :returns: a nested dict with the relevant information.
        """
        from aiida.orm import Comment, Computer, Group, Log, Node, QueryBuilder, User

        data = {}

        query_user = QueryBuilder(self).append(User, project=['email'])
        data['Users'] = {'count': query_user.count()}
        if detailed:
            data['Users']['emails'] = sorted({email for email, in query_user.iterall() if email is not None})

        query_comp = QueryBuilder(self).append(Computer, project=['label'])
        data['Computers'] = {'count': query_comp.count()}
        if detailed:
            data['Computers']['labels'] = sorted({comp for comp, in query_comp.iterall() if comp is not None})

        count = QueryBuilder(self).append(Node).count()
        data['Nodes'] = {'count': count}
        if detailed:
            node_types = sorted({
                typ for typ, in QueryBuilder(self).append(Node, project=['node_type']).iterall() if typ is not None
            })
            data['Nodes']['node_types'] = node_types
            process_types = sorted({
                typ for typ, in QueryBuilder(self).append(Node, project=['process_type']).iterall() if typ is not None
            })
            data['Nodes']['process_types'] = [p for p in process_types if p]

        query_group = QueryBuilder(self).append(Group, project=['type_string'])
        data['Groups'] = {'count': query_group.count()}
        if detailed:
            data['Groups']['type_strings'] = sorted({typ for typ, in query_group.iterall() if typ is not None})

        count = QueryBuilder(self).append(Comment).count()
        data['Comments'] = {'count': count}

        count = QueryBuilder(self).append(Log).count()
        data['Logs'] = {'count': count}

        count = QueryBuilder(self).append(entity_type='link').count()
        data['Links'] = {'count': count}

        return data



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Versioning and migration implementation for the sqlite_zip format."""
import contextlib
from datetime import datetime
import json
import os
from pathlib import Path
import shutil
import tarfile
import tempfile
from typing import Any, Dict, Iterator, List, Optional, Union
import zipfile

from alembic.command import upgrade
from alembic.config import Config
from alembic.runtime.environment import EnvironmentContext
from alembic.runtime.migration import MigrationContext, MigrationInfo
from alembic.script import ScriptDirectory
from archive_path import ZipPath, extract_file_in_zip, open_file_in_tar, open_file_in_zip

from aiida.common.exceptions import CorruptStorage, IncompatibleStorageSchema, StorageMigrationError
from aiida.common.progress_reporter import get_progress_reporter
from aiida.storage.log import MIGRATE_LOGGER

from .migrations.legacy import FINAL_LEGACY_VERSION, LEGACY_MIGRATE_FUNCTIONS
from .migrations.legacy_to_main import LEGACY_TO_MAIN_REVISION, perform_v1_migration
from .migrations.utils import copy_tar_to_zip, copy_zip_to_zip, update_metadata
from .utils import DB_FILENAME, META_FILENAME, REPO_FOLDER, create_sqla_engine, extract_metadata, read_version


def get_schema_version_head() -> str:
    """Return the head schema version for this storage, i.e. the latest schema this storage can be migrated to."""
    return _alembic_script().revision_map.get_current_head('main') or ''


def list_versions() -> List[str]:
    """Return all available schema versions (oldest to latest)."""
    legacy_versions = list(LEGACY_MIGRATE_FUNCTIONS) + [FINAL_LEGACY_VERSION]
    alembic_versions = [entry.revision for entry in reversed(list(_alembic_script().walk_revisions()))]
    return legacy_versions + alembic_versions


def validate_storage(inpath: Path) -> None:
    """Validate that the storage is at the head version.

    :raises: :class:`aiida.common.exceptions.UnreachableStorage` if the file does not exist
    :raises: :class:`aiida.common.exceptions.CorruptStorage`
        if the version cannot be read from the storage.
    :raises: :class:`aiida.common.exceptions.IncompatibleStorageSchema`
        if the storage is not compatible with the code API.
    """
    schema_version_code = get_schema_version_head()
    schema_version_archive = read_version(inpath)
    if schema_version_archive != schema_version_code:
        raise IncompatibleStorageSchema(
            f'Archive schema version `{schema_version_archive}` '
            f'is incompatible with the required schema version `{schema_version_code}`. '
            'To migrate the archive schema version to the current one, '
            f'run the following command: verdi archive migrate {str(inpath)!r}'
        )


def migrate(  # pylint: disable=too-many-branches,too-many-statements,too-many-locals
    inpath: Union[str, Path],
    outpath: Union[str, Path],
    version: str,
    *,
    force: bool = False,
    compression: int = 6
) -> None:
    """Migrate an `sqlite_zip` storage file to a specific version.

    Historically, this format could be a zip or a tar file,
    contained the database as a bespoke JSON format, and the repository files in the "legacy" per-node format.
    For these versions, we first migrate the JSON database to the final legacy schema,
    then we convert this file to the SQLite database, whilst sequentially migrating the repository files.

    Once any legacy migrations have been performed, we can then migrate the SQLite database to the final schema,
    using alembic.

    Note that, to minimise disk space usage, we never fully extract/uncompress the input file
    (except when migrating from a legacy tar file, whereby we cannot extract individual files):

    1. The sqlite database is extracted to a temporary location and migrated
    2. A new zip file is opened, within a temporary folder
    3. The repository files are "streamed" directly between the input file and the new zip file
    4. The sqlite database and metadata JSON are written to the new zip file
    5. The new zip file is closed (which writes its final central directory)
    6. The new zip file is moved to the output location, removing any existing file if `force=True`

    :param path: Path to the file
    :param outpath: Path to output the migrated file
    :param version: Target version
    :param force: If True, overwrite the output file if it exists
    :param compression: Compression level for the output file
    """
    inpath = Path(inpath)
    outpath = Path(outpath)

    # halt immediately, if we could not write to the output file
    if outpath.exists() and not force:
        raise StorageMigrationError('Output path already exists and force=False')
    if outpath.exists() and not outpath.is_file():
        raise StorageMigrationError('Existing output path is not a file')

    # the file should be either a tar (legacy only) or zip file
    if tarfile.is_tarfile(str(inpath)):
        is_tar = True
    elif zipfile.is_zipfile(str(inpath)):
        is_tar = False
    else:
        raise CorruptStorage(f'The input file is neither a tar nor a zip file: {inpath}')

    # read the metadata.json which should always be present
    metadata = extract_metadata(inpath, search_limit=None)

    # obtain the current version from the metadata
    if 'export_version' not in metadata:
        raise CorruptStorage('No export_version found in metadata')
    current_version = metadata['export_version']
    # update the modified time of the file and the compression
    metadata['mtime'] = datetime.now().isoformat()
    metadata['compression'] = compression

    # check versions are valid
    # versions 0.1, 0.2, 0.3 are no longer supported,
    # since 0.3 -> 0.4 requires costly migrations of repo files (you would need to unpack all of them)
    if current_version in ('0.1', '0.2', '0.3') or version in ('0.1', '0.2', '0.3'):
        raise StorageMigrationError(
            f"Legacy migration from '{current_version}' -> '{version}' is not supported in aiida-core v2. "
            'First migrate them to the latest version in aiida-core v1.'
        )
    all_versions = list_versions()
    if current_version not in all_versions:
        raise StorageMigrationError(f"Unknown current version '{current_version}'")
    if version not in all_versions:
        raise StorageMigrationError(f"Unknown target version '{version}'")

    # if we are already at the desired version, then no migration is required, so simply copy the file if necessary
    if current_version == version:
        if inpath != outpath:
            if outpath.exists() and force:
                outpath.unlink()
            shutil.copyfile(inpath, outpath)
        return

    # if the archive is a "legacy" format, i.e. has a data.json file, migrate it to the target/final legacy schema
    data: Optional[Dict[str, Any]] = None
    if current_version in LEGACY_MIGRATE_FUNCTIONS:
        MIGRATE_LOGGER.report(f'Legacy migrations required from {"tar" if is_tar else "zip"} format')
        MIGRATE_LOGGER.report('Extracting data.json ...')
        # read the data.json file
        data = _read_json(inpath, 'data.json', is_tar)
        to_version = FINAL_LEGACY_VERSION if version not in LEGACY_MIGRATE_FUNCTIONS else version
        current_version = _perform_legacy_migrations(current_version, to_version, metadata, data)

    # if we are now at the target version, then write the updated files to a new zip file and exit
    if current_version == version:
        # create new legacy archive with updated metadata & data
        def path_callback(inpath, outpath) -> bool:
            if inpath.name == 'metadata.json':
                outpath.write_text(json.dumps(metadata))
                return True
            if inpath.name == 'data.json':
                outpath.write_text(json.dumps(data))
                return True
            return False

        func = copy_tar_to_zip if is_tar else copy_zip_to_zip

        func(
            inpath,
            outpath,
            path_callback,
            overwrite=force,
            compression=compression,
            title='Writing migrated legacy archive',
            info_order=('metadata.json', 'data.json')
        )
        return

    # open the temporary directory, to perform further migrations
    with tempfile.TemporaryDirectory() as tmpdirname:

        # open the new zip file, within which to write the migrated content
        new_zip_path = Path(tmpdirname) / 'new.zip'
        central_dir: Dict[str, Any] = {}
        with ZipPath(
            new_zip_path,
            mode='w',
            compresslevel=compression,
            name_to_info=central_dir,
            # this ensures that the metadata and database files are written above the repository files,
            # in in the central directory, so that they can be accessed easily
            info_order=(META_FILENAME, DB_FILENAME)
        ) as new_zip:

            written_repo = False
            if current_version == FINAL_LEGACY_VERSION:
                # migrate from the legacy format,
                # streaming the repository files directly to the new zip file
                MIGRATE_LOGGER.report(
                    f'legacy {FINAL_LEGACY_VERSION!r} -> {LEGACY_TO_MAIN_REVISION!r} conversion required'
                )
                if data is None:
                    MIGRATE_LOGGER.report('Extracting data.json ...')
                    data = _read_json(inpath, 'data.json', is_tar)
                db_path = perform_v1_migration(inpath, Path(tmpdirname), new_zip, central_dir, is_tar, metadata, data)
                # the migration includes adding the repository files to the new zip file
                written_repo = True
                current_version = LEGACY_TO_MAIN_REVISION
            else:
                if is_tar:
                    raise CorruptStorage('Tar files are not supported for this format')
                # extract the sqlite database, for alembic migrations
                db_path = Path(tmpdirname) / DB_FILENAME
                with db_path.open('wb') as handle:
                    try:
                        extract_file_in_zip(inpath, DB_FILENAME, handle)
                    except Exception as exc:
                        raise CorruptStorage(f'database could not be read: {exc}') from exc

            # perform alembic migrations
            # note, we do this before writing the repository files (unless a legacy migration),
            # so that we don't waste time doing that (which could be slow), only for alembic to fail
            if current_version != version:
                MIGRATE_LOGGER.report('Performing SQLite migrations:')
                with _migration_context(db_path) as context:
                    assert context.script is not None
                    context.stamp(context.script, current_version)
                    context.connection.commit()  # type: ignore
                # see https://alembic.sqlalchemy.org/en/latest/batch.html#dealing-with-referencing-foreign-keys
                # for why we do not enforce foreign keys here
                with _alembic_connect(db_path, enforce_foreign_keys=False) as config:
                    upgrade(config, version)
                update_metadata(metadata, version)

            if not written_repo:
                # stream the repository files directly to the new zip file
                with ZipPath(inpath, mode='r') as old_zip:
                    length = sum(1 for _ in old_zip.glob('**/*', include_virtual=False))
                    title = 'Copying repository files'
                    with get_progress_reporter()(desc=title, total=length) as progress:
                        for subpath in old_zip.glob('**/*', include_virtual=False):
                            new_path_sub = new_zip.joinpath(subpath.at)
                            if subpath.parts[0] == REPO_FOLDER:
                                if subpath.is_dir():
                                    new_path_sub.mkdir(exist_ok=True)
                                else:
                                    new_path_sub.putfile(subpath)
                            progress.update()

            MIGRATE_LOGGER.report('Finalising the migration ...')

            # write the final database file to the new zip file
            with db_path.open('rb') as handle:
                with (new_zip / DB_FILENAME).open(mode='wb') as handle2:
                    shutil.copyfileobj(handle, handle2)

            # write the final metadata.json file to the new zip file
            (new_zip / META_FILENAME).write_text(json.dumps(metadata))

            # on exiting the the ZipPath context, the zip file is closed and the central directory written

        # move the new zip file to the final location
        if outpath.exists() and force:
            outpath.unlink()
        shutil.move(new_zip_path, outpath)  # type: ignore[arg-type]


def _read_json(inpath: Path, filename: str, is_tar: bool) -> Dict[str, Any]:
    """Read a JSON file from the archive."""
    if is_tar:
        with open_file_in_tar(inpath, filename) as handle:
            data = json.load(handle)
    else:
        with open_file_in_zip(inpath, filename) as handle:
            data = json.load(handle)
    return data


def _perform_legacy_migrations(current_version: str, to_version: str, metadata: dict, data: dict) -> str:
    """Perform legacy migrations from the current version to the desired version.

    Legacy archives use the old ``data.json`` format for storing the database.
    These migrations simply manipulate the metadata and data in-place.

    :param current_version: current version of the archive
    :param to_version: version to migrate to
    :param metadata: the metadata to migrate
    :param data: the data to migrate
    :return: the new version of the archive
    """
    # compute the migration pathway
    prev_version = current_version
    pathway: List[str] = []
    while prev_version != to_version:
        if prev_version not in LEGACY_MIGRATE_FUNCTIONS:
            raise StorageMigrationError(f"No migration pathway available for '{current_version}' to '{to_version}'")
        if prev_version in pathway:
            raise StorageMigrationError(
                f'cyclic migration pathway encountered: {" -> ".join(pathway + [prev_version])}'
            )
        pathway.append(prev_version)
        prev_version = LEGACY_MIGRATE_FUNCTIONS[prev_version][0]

    if not pathway:
        MIGRATE_LOGGER.report('No migration required')
        return to_version

    MIGRATE_LOGGER.report('Legacy migration pathway: %s', ' -> '.join(pathway + [to_version]))

    with get_progress_reporter()(total=len(pathway), desc='Performing migrations: ') as progress:
        for from_version in pathway:
            to_version = LEGACY_MIGRATE_FUNCTIONS[from_version][0]
            progress.set_description_str(f'Performing migrations: {from_version} -> {to_version}', refresh=True)
            LEGACY_MIGRATE_FUNCTIONS[from_version][1](metadata, data)
            progress.update()

    return to_version


def _alembic_config() -> Config:
    """Return an instance of an Alembic `Config`."""
    config = Config()
    config.set_main_option('script_location', str(Path(os.path.realpath(__file__)).parent / 'migrations'))
    return config


def _alembic_script() -> ScriptDirectory:
    """Return an instance of an Alembic `ScriptDirectory`."""
    return ScriptDirectory.from_config(_alembic_config())


@contextlib.contextmanager
def _alembic_connect(db_path: Path, enforce_foreign_keys=True) -> Iterator[Config]:
    """Context manager to return an instance of an Alembic configuration.

    The profiles's database connection is added in the `attributes` property, through which it can then also be
    retrieved, also in the `env.py` file, which is run when the database is migrated.
    """
    with create_sqla_engine(db_path, enforce_foreign_keys=enforce_foreign_keys).connect() as connection:
        config = _alembic_config()
        config.attributes['connection'] = connection  # pylint: disable=unsupported-assignment-operation

        def _callback(step: MigrationInfo, **kwargs):  # pylint: disable=unused-argument
            """Callback to be called after a migration step is executed."""
            from_rev = step.down_revision_ids[0] if step.down_revision_ids else '<base>'
            MIGRATE_LOGGER.report(f'- {from_rev} -> {step.up_revision_id}')

        config.attributes['on_version_apply'] = _callback  # pylint: disable=unsupported-assignment-operation

        yield config


@contextlib.contextmanager
def _migration_context(db_path: Path) -> Iterator[MigrationContext]:
    """Context manager to return an instance of an Alembic migration context.

    This migration context will have been configured with the current database connection, which allows this context
    to be used to inspect the contents of the database, such as the current revision.
    """
    with _alembic_connect(db_path) as config:
        script = ScriptDirectory.from_config(config)
        with EnvironmentContext(config, script) as context:
            context.configure(context.config.attributes['connection'])
            yield context.get_context()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi data array` command."""

from aiida.cmdline.commands.cmd_data import verdi_data
from aiida.cmdline.params import arguments, options, types


@verdi_data.group('array')
def array():
    """Manipulate ArrayData objects (numpy arrays)."""


@array.command('show')
@arguments.DATA(type=types.DataParamType(sub_classes=('aiida.data:core.array',)))
@options.DICT_FORMAT()
def array_show(data, fmt):
    """Visualize ArrayData objects."""
    from aiida.cmdline.utils.echo import echo_dictionary

    for node in data:
        the_dict = {}
        for arrayname in node.get_arraynames():
            the_dict[arrayname] = node.get_array(arrayname).tolist()
        echo_dictionary(the_dict, fmt)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Mixin classes for ORM classes."""
import inspect
import io
import tempfile

from aiida.common import exceptions
from aiida.common.lang import classproperty, override


class FunctionCalculationMixin:
    """
    This mixin should be used for ProcessNode subclasses that are used to record the execution
    of a python function. For example the process nodes that are used for a function that
    was wrapped by the `workfunction` or `calcfunction` function decorators. The `store_source_info`
    method can then be called with the wrapped function to store information about that function
    in the calculation node through the inspect module. Various property getters are defined to
    later retrieve that information from the node
    """

    FUNCTION_NAME_KEY = 'function_name'
    FUNCTION_NAMESPACE_KEY = 'function_namespace'
    FUNCTION_STARTING_LINE_KEY = 'function_starting_line_number'
    FUNCTION_SOURCE_FILE_PATH = 'source_file'

    def store_source_info(self, func):
        """
        Retrieve source information about the wrapped function `func` through the inspect module,
        and store it in the attributes and repository of the node. The function name, namespace
        and the starting line number in the source file will be stored in the attributes. The
        source file itself will be copied into the repository

        :param func: the function to inspect and whose information to store in the node
        """
        self._set_function_name(func.__name__)

        try:
            _, starting_line_number = inspect.getsourcelines(func)
            self._set_function_starting_line_number(starting_line_number)
        except (IOError, OSError):
            pass

        try:
            self._set_function_namespace(func.__globals__['__name__'])
        except Exception:  # pylint: disable=broad-except
            pass

        try:
            source_file_path = inspect.getsourcefile(func)
            with open(source_file_path, 'rb') as handle:
                self.put_object_from_filelike(handle, self.FUNCTION_SOURCE_FILE_PATH)
        except (IOError, OSError):
            pass

    @property
    def function_name(self):
        """Return the function name of the wrapped function.

        :returns: the function name or None
        """
        return self.get_attribute(self.FUNCTION_NAME_KEY, None)

    def _set_function_name(self, function_name):
        """Set the function name of the wrapped function.

        :param function_name: the function name
        """
        self.set_attribute(self.FUNCTION_NAME_KEY, function_name)

    @property
    def function_namespace(self):
        """Return the function namespace of the wrapped function.

        :returns: the function namespace or None
        """
        return self.get_attribute(self.FUNCTION_NAMESPACE_KEY, None)

    def _set_function_namespace(self, function_namespace):
        """Set the function namespace of the wrapped function.

        :param function_namespace: the function namespace
        """
        self.set_attribute(self.FUNCTION_NAMESPACE_KEY, function_namespace)

    @property
    def function_starting_line_number(self):
        """Return the starting line number of the wrapped function in its source file.

        :returns: the starting line number or None
        """
        return self.get_attribute(self.FUNCTION_STARTING_LINE_KEY, None)

    def _set_function_starting_line_number(self, function_starting_line_number):
        """Set the starting line number of the wrapped function in its source file.

        :param function_starting_line_number: the starting line number
        """
        self.set_attribute(self.FUNCTION_STARTING_LINE_KEY, function_starting_line_number)

    def get_function_source_code(self):
        """Return the absolute path to the source file in the repository.

        :returns: the absolute path of the source file in the repository, or None if it does not exist
        """
        return self.get_object_content(self.FUNCTION_SOURCE_FILE_PATH)


class Sealable:
    """Mixin to mark a Node as `sealable`."""
    # pylint: disable=no-member,unsupported-membership-test

    SEALED_KEY = 'sealed'

    @classproperty
    def _updatable_attributes(cls):  # pylint: disable=no-self-argument
        return (cls.SEALED_KEY,)

    def check_mutability(self):
        """Check if the node is mutable.

        :raises `~aiida.common.exceptions.ModificationNotAllowed`: when the node is sealed and therefore immutable.
        """
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('the node is sealed and therefore the repository is immutable.')

    def validate_incoming(self, source, link_type, link_label):
        """Validate adding a link of the given type from a given node to ourself.

        Adding an incoming link to a sealed node is forbidden.

        :param source: the node from which the link is coming
        :param link_type: the link type
        :param link_label: the link label
        :raise aiida.common.ModificationNotAllowed: if the target node (self) is sealed
        """
        if self.is_sealed:
            raise exceptions.ModificationNotAllowed('Cannot add a link to a sealed node')

        super().validate_incoming(source, link_type=link_type, link_label=link_label)

    def validate_outgoing(self, target, link_type, link_label):
        """Validate adding a link of the given type from ourself to a given node.

        Adding an outgoing link from a sealed node is forbidden.

        :param target: the node to which the link is going
        :param link_type: the link type
        :param link_label: the link label
        :raise aiida.common.ModificationNotAllowed: if the source node (self) is sealed
        """
        if self.is_sealed:
            raise exceptions.ModificationNotAllowed('Cannot add a link from a sealed node')

        super().validate_outgoing(target, link_type=link_type, link_label=link_label)

    @property
    def is_sealed(self):
        """Returns whether the node is sealed, i.e. whether the sealed attribute has been set to True."""
        return self.get_attribute(self.SEALED_KEY, False)

    def seal(self):
        """Seal the node by setting the sealed attribute to True."""
        if not self.is_sealed:
            self.set_attribute(self.SEALED_KEY, True)

    @override
    def set_attribute(self, key, value):
        """Set an attribute to the given value.

        :param key: name of the attribute
        :param value: value of the attribute
        :raise aiida.common.exceptions.ModificationNotAllowed: if the node is already sealed or if the node
            is already stored and the attribute is not updatable.
        """
        if self.is_sealed:
            raise exceptions.ModificationNotAllowed('attributes of a sealed node are immutable')

        if self.is_stored and key not in self._updatable_attributes:  # pylint: disable=unsupported-membership-test
            raise exceptions.ModificationNotAllowed(f'`{key}` is not an updatable attribute')

        self.backend_entity.set_attribute(key, value)

    @override
    def delete_attribute(self, key):
        """Delete an attribute.

        :param key: name of the attribute
        :raises AttributeError: if the attribute does not exist
        :raise aiida.common.exceptions.ModificationNotAllowed: if the node is already sealed or if the node
            is already stored and the attribute is not updatable.
        """
        if self.is_sealed:
            raise exceptions.ModificationNotAllowed('attributes of a sealed node are immutable')

        if self.is_stored and key not in self._updatable_attributes:  # pylint: disable=unsupported-membership-test
            raise exceptions.ModificationNotAllowed(f'`{key}` is not an updatable attribute')

        self.backend_entity.delete_attribute(key)

    @override
    def put_object_from_filelike(self, handle: io.BufferedReader, path: str):
        """Store the byte contents of a file in the repository.

        :param handle: filelike object with the byte content to be stored.
        :param path: the relative path where to store the object in the repository.
        :raises TypeError: if the path is not a string and relative path.
        :raises aiida.common.exceptions.ModificationNotAllowed: when the node is sealed and therefore immutable.
        """
        self.check_mutability()

        if isinstance(handle, io.StringIO):
            handle = io.BytesIO(handle.read().encode('utf-8'))

        if isinstance(handle, tempfile._TemporaryFileWrapper):  # pylint: disable=protected-access
            if 'b' in handle.file.mode:
                handle = io.BytesIO(handle.read())
            else:
                handle = io.BytesIO(handle.read().encode('utf-8'))

        self._repository.put_object_from_filelike(handle, path)
        self._update_repository_metadata()

    @override
    def put_object_from_file(self, filepath: str, path: str):
        """Store a new object under `path` with contents of the file located at `filepath` on the local file system.

        :param filepath: absolute path of file whose contents to copy to the repository
        :param path: the relative path where to store the object in the repository.
        :raises TypeError: if the path is not a string and relative path, or the handle is not a byte stream.
        :raises aiida.common.exceptions.ModificationNotAllowed: when the node is sealed and therefore immutable.
        """
        self.check_mutability()
        self._repository.put_object_from_file(filepath, path)
        self._update_repository_metadata()

    @override
    def put_object_from_tree(self, filepath: str, path: str = None):
        """Store the entire contents of `filepath` on the local file system in the repository with under given `path`.

        :param filepath: absolute path of the directory whose contents to copy to the repository.
        :param path: the relative path where to store the objects in the repository.
        :raises TypeError: if the path is not a string and relative path.
        :raises aiida.common.exceptions.ModificationNotAllowed: when the node is sealed and therefore immutable.
        """
        self.check_mutability()
        self._repository.put_object_from_tree(filepath, path)
        self._update_repository_metadata()

    @override
    def delete_object(self, path: str):
        """Delete the object from the repository.

        :param key: fully qualified identifier for the object within the repository.
        :raises TypeError: if the path is not a string and relative path.
        :raises FileNotFoundError: if the file does not exist.
        :raises IsADirectoryError: if the object is a directory and not a file.
        :raises OSError: if the file could not be deleted.
        :raises aiida.common.exceptions.ModificationNotAllowed: when the node is sealed and therefore immutable.
        """
        self.check_mutability()
        self._repository.delete_object(path)
        self._update_repository_metadata()
